from __future__ import annotations

import time
from enum import Enum

from mypy_boto3_ec2.client import EC2Client
from mypy_boto3_ec2.type_defs import (
    DescribeInstancesRequestTypeDef,
    DescribeInstanceStatusRequestTypeDef,
    InstanceStateChangeTypeDef,
    InstanceStateTypeDef,
    InstanceStatusTypeDef,
    RebootInstancesRequestTypeDef,
    StartInstancesRequestTypeDef,
    StopInstancesRequestTypeDef,
)
from rich import console as rich_console


class Ec2InstanceState(str, Enum):
    """State of the EC2 instance."""

    PENDING = "pending"
    RUNNING = "running"
    SHUTTING_DOWN = "shutting-down"
    TERMINATED = "terminated"
    STOPPING = "stopping"
    STOPPED = "stopped"

    @classmethod
    def from_name(cls, state_name: str) -> Ec2InstanceState:
        """Return the enum value, ignoring case.

        Raises:
            ValueError if no matching enum value is found.
        """
        name_lower = state_name.lower()
        for state in cls:
            if state.value == name_lower:
                return state
        raise ValueError(f"No Ec2InstanceState found for '{state_name}'")

    @classmethod
    def from_state_response(cls, instance_state: InstanceStateTypeDef) -> Ec2InstanceState:
        """Return the enum value.

        Raises:
            ValueError if no matching code or name is found.
        """
        state_code = instance_state.get("Code")
        state_name = instance_state.get("Name")

        if state_code is not None:
            try:
                return _INSTANCE_REVERSE_CODE_MAP[state_code]
            except KeyError as e:
                raise ValueError(f"Unknown state code: {state_code}") from e

        if state_name is not None:
            return Ec2InstanceState.from_name(state_name)

        raise ValueError(f"Neither code not name found in instance state: {instance_state}")

    def get_code(self) -> int:
        """Return the corresponding instance state code."""
        return _INSTANCE_CODE_MAP[self]

    def is_terminal(self) -> bool:
        """Return True if the instance state is not transitory."""
        return self in [
            Ec2InstanceState.RUNNING,
            Ec2InstanceState.TERMINATED,
            Ec2InstanceState.STOPPED,
        ]

    def is_startable(self) -> bool:
        """Return True if the instance can be started."""
        return self == Ec2InstanceState.STOPPED

    def is_stoppable(self) -> bool:
        """Return True if the instance can be stopped."""
        return self == Ec2InstanceState.RUNNING


# see https://docs.aws.amazon.com/AWSEC2/latest/APIReference/API_InstanceState.html
_INSTANCE_CODE_MAP: dict[Ec2InstanceState, int] = {
    Ec2InstanceState.PENDING: 0,
    Ec2InstanceState.RUNNING: 16,
    Ec2InstanceState.SHUTTING_DOWN: 32,
    Ec2InstanceState.TERMINATED: 48,
    Ec2InstanceState.STOPPING: 64,
    Ec2InstanceState.STOPPED: 80,
}
_INSTANCE_REVERSE_CODE_MAP: dict[int, Ec2InstanceState] = {v: k for k, v in _INSTANCE_CODE_MAP.items()}


def describe_instance_status(
    ec2_client: EC2Client, instance_id: str, check_status_first: bool = True
) -> InstanceStatusTypeDef:
    """Call one of the EC2 describe-instance APIs, return the InstanceStatus.

    Raises:
        ValueError if the instance is not found.
    """

    # first try calling EC2:DescribeInstanceStatuses
    # this will only surface running instances
    if check_status_first:
        request: DescribeInstanceStatusRequestTypeDef = {"InstanceIds": [instance_id]}
        response = ec2_client.describe_instance_status(**request)

        instance_statuses = response["InstanceStatuses"]

        if instance_statuses:
            return instance_statuses[0]

    # second try calling EC2:DescribeInstance directly
    full_describe_request: DescribeInstancesRequestTypeDef = {"InstanceIds": [instance_id]}
    full_response = ec2_client.describe_instances(**full_describe_request)
    reservations = full_response["Reservations"]

    if not reservations:
        raise ValueError("Instance not found: no reservation.")

    instances = reservations[0].get("Instances", [])

    if not instances:
        raise ValueError("Instance not found in reservation")
    instance = instances[0]

    instance_status: InstanceStatusTypeDef = {"InstanceState": instance.get("State", {})}
    return instance_status


def poll_for_instance_status(
    ec2_client: EC2Client,
    console: rich_console.Console,
    instance_id: str,
    desired_state: Ec2InstanceState,
    timeout_seconds: int = 60,
    wait_after_seconds: int = 2,
    poll_interval_seconds: int = 5,
) -> InstanceStatusTypeDef:
    """Synchronously poll EC2:GetInstanceStatus until the instance reaches a terminal state.

    Raises:
        ValueError if the instance reaches a terminal state that is not the desired state
    """
    # allow instance to change state
    if wait_after_seconds > 0:
        time.sleep(wait_after_seconds)

    start_time = time.time()
    while True:
        response = describe_instance_status(ec2_client, instance_id=instance_id, check_status_first=False)
        state_response = response.get("InstanceState", {})
        state = Ec2InstanceState.from_state_response(state_response)
        curr_time = time.time()

        if state == desired_state:
            console.print(f"Instance reached desired state: '{desired_state.value}'")
            return response
        elif state.is_terminal():
            raise ValueError(f"Unexpected terminal state for instance '{instance_id}': '{state.value}'")
        elif curr_time - start_time > timeout_seconds:
            raise TimeoutError(f"Timed out polling state of instance '{instance_id}', end state '{state.value}'")
        else:
            console.print(f"Instance state is '{state.value}', waiting for '{desired_state.value}'...")
            time.sleep(poll_interval_seconds)


def start_instance(ec2_client: EC2Client, instance_id: str) -> InstanceStateChangeTypeDef:
    """Call EC2:StartInstance, return the InstanceStateChange.

    Raises:
        ValueError if the instance is not found.
    """

    request: StartInstancesRequestTypeDef = {"InstanceIds": [instance_id]}
    response = ec2_client.start_instances(**request)

    instance_state_changes = response["StartingInstances"]

    if not instance_state_changes:
        raise ValueError("Instance ID not found.")

    return instance_state_changes[0]


def stop_instance(ec2_client: EC2Client, instance_id: str) -> InstanceStateChangeTypeDef:
    """Call EC2:StopInstance, return the InstanceStateChange.

    Raises:
        ValueError if the instance is not found.
    """

    request: StopInstancesRequestTypeDef = {"InstanceIds": [instance_id]}
    response = ec2_client.stop_instances(**request)

    instance_state_changes = response["StoppingInstances"]

    if not instance_state_changes:
        raise ValueError("Instance ID not found.")

    return instance_state_changes[0]


def restart_instance(ec2_client: EC2Client, instance_id: str) -> None:
    """Call EC2:RebootInstance."""

    request: RebootInstancesRequestTypeDef = {"InstanceIds": [instance_id]}
    ec2_client.reboot_instances(**request)



import unittest
from pathlib import Path
from unittest.mock import Mock, patch

from jupyter_deploy.engine.enum import EngineType
from jupyter_deploy.engine.terraform.tf_up import TerraformUpHandler


class TestTerraformUpHandler(unittest.TestCase):
    """Test cases for the TerraformUpHandler class."""

    def test_init_sets_attributes(self) -> None:
        project_path = Path("/mock/project")
        handler = TerraformUpHandler(project_path=project_path)

        self.assertEqual(handler.project_path, project_path)
        self.assertEqual(handler.engine, EngineType.TERRAFORM)

    def test_get_default_config_filename_returns_terraform_default(self) -> None:
        project_path = Path("/mock/project")
        handler = TerraformUpHandler(project_path=project_path)

        result = handler.get_default_config_filename()

        self.assertEqual(result, "jdout-tfplan")

    @patch("jupyter_deploy.engine.terraform.tf_up.cmd_utils")
    @patch("jupyter_deploy.engine.terraform.tf_up.rich_console")
    def test_apply_success(self, mock_console: Mock, mock_cmd_utils: Mock) -> None:
        path = Path("/mock/path")
        project_path = Path("/mock/project")
        engine_path = project_path / "engine"
        handler = TerraformUpHandler(project_path=project_path)

        mock_console_instance = Mock()
        mock_console.Console.return_value = mock_console_instance

        mock_cmd_utils.run_cmd_and_pipe_to_terminal.return_value = (0, False)

        handler.apply(path)

        mock_cmd_utils.run_cmd_and_pipe_to_terminal.assert_called_once_with(
            ["terraform", "apply", "/mock/path"], exec_dir=engine_path
        )
        mock_console_instance.print.assert_called_once()
        self.assertTrue(mock_console_instance.print.call_args[0][0].lower().find("success") >= 0)

    @patch("jupyter_deploy.engine.terraform.tf_up.cmd_utils")
    @patch("jupyter_deploy.engine.terraform.tf_up.rich_console")
    def test_apply_handles_error(self, mock_console: Mock, mock_cmd_utils: Mock) -> None:
        path = Path("/mock/path")
        project_path = Path("/mock/project")
        engine_path = project_path / "engine"
        handler = TerraformUpHandler(project_path=project_path)

        mock_console_instance = Mock()
        mock_console.Console.return_value = mock_console_instance

        mock_cmd_utils.run_cmd_and_pipe_to_terminal.return_value = (1, False)

        handler.apply(path)

        mock_cmd_utils.run_cmd_and_pipe_to_terminal.assert_called_once_with(
            ["terraform", "apply", "/mock/path"], exec_dir=engine_path
        )
        mock_console_instance.print.assert_called_once()
        self.assertTrue(mock_console_instance.print.call_args[0][0].lower().find("error") >= 0)

    @patch("jupyter_deploy.engine.terraform.tf_up.cmd_utils")
    @patch("jupyter_deploy.engine.terraform.tf_up.rich_console")
    def test_apply_handles_timeout(self, mock_console: Mock, mock_cmd_utils: Mock) -> None:
        path = Path("/mock/path")
        project_path = Path("/mock/project")
        engine_path = project_path / "engine"
        handler = TerraformUpHandler(project_path=project_path)

        mock_console_instance = Mock()
        mock_console.Console.return_value = mock_console_instance

        mock_cmd_utils.run_cmd_and_pipe_to_terminal.return_value = (0, True)

        handler.apply(path)

        mock_cmd_utils.run_cmd_and_pipe_to_terminal.assert_called_once_with(
            ["terraform", "apply", "/mock/path"], exec_dir=engine_path
        )
        mock_console_instance.print.assert_called_once()
        self.assertTrue(mock_console_instance.print.call_args[0][0].lower().find("error") >= 0)

    @patch("jupyter_deploy.engine.terraform.tf_up.cmd_utils")
    def test_apply_propagates_exceptions(self, mock_cmd_utils: Mock) -> None:
        path = Path("/mock/path")
        project_path = Path("/mock/project")
        handler = TerraformUpHandler(project_path=project_path)

        mock_cmd_utils.run_cmd_and_pipe_to_terminal.side_effect = Exception("Command failed")

        with self.assertRaises(Exception) as context:
            handler.apply(path)

        self.assertEqual(str(context.exception), "Command failed")
        mock_cmd_utils.run_cmd_and_pipe_to_terminal.assert_called_once()

    @patch("jupyter_deploy.engine.terraform.tf_up.cmd_utils")
    @patch("jupyter_deploy.engine.terraform.tf_up.rich_console")
    def test_apply_with_auto_approve(self, mock_console: Mock, mock_cmd_utils: Mock) -> None:
        """Test that auto_approve flag is properly passed to terraform command."""
        path = Path("/mock/path")
        project_path = Path("/mock/project")
        handler = TerraformUpHandler(project_path=project_path)

        mock_console_instance = Mock()
        mock_console.Console.return_value = mock_console_instance
        mock_cmd_utils.run_cmd_and_pipe_to_terminal.return_value = (0, False)

        handler.apply(path, auto_approve=True)

        mock_cmd_utils.run_cmd_and_pipe_to_terminal.assert_called_once()
        cmd_args = mock_cmd_utils.run_cmd_and_pipe_to_terminal.call_args[0][0]
        self.assertIn("-auto-approve", cmd_args)



import unittest
from typing import Any
from unittest.mock import Mock, patch

from parameterized import parameterized  # type: ignore

from jupyter_deploy.engine.terraform.tf_types import TerraformType
from jupyter_deploy.engine.terraform.tf_vardefs import (
    TerraformBoolVariableDefinition,
    TerraformListOfMapStrVariableDefinition,
    TerraformListOfStrVariableDefinition,
    TerraformMapOfStrVariableDefinition,
    TerraformNumberVariableDefinition,
    TerraformStrVariableDefinition,
    create_tf_variable_definition,
    to_tf_var_option,
)
from jupyter_deploy.engine.vardefs import (
    AnyNumericTemplateVariableDefinition,
    BoolTemplateVariableDefinition,
    DictStrTemplateVariableDefinition,
    FloatTemplateVariableDefinition,
    IntTemplateVariableDefinition,
    ListMapStrTemplateVariableDefinition,
    ListStrTemplateVariableDefinition,
    StrTemplateVariableDefinition,
)


class TestTerraformTemplateVariableClasses(unittest.TestCase):
    def test_str_var_instantiable_and_converts_to_str_template(self) -> None:
        # Arrange
        tf_var = TerraformStrVariableDefinition(
            variable_name="test_var",
            description="Test variable",
            default="default_value",
        )

        # Act
        template_var = tf_var.to_template_definition()

        # Assert
        self.assertIsInstance(template_var, StrTemplateVariableDefinition)
        self.assertEqual(template_var.variable_name, "test_var")
        self.assertEqual(template_var.description, "Test variable")
        self.assertEqual(template_var.default, "default_value")

    def test_number_var_instantiable_with_int_and_converts_to_int_vardef(self) -> None:
        # Arrange
        tf_var = TerraformNumberVariableDefinition(
            variable_name="test_int_var",
            description="Test integer variable",
            default=42,
        )

        # Act
        template_var = tf_var.to_template_definition()

        # Assert
        self.assertIsInstance(template_var, IntTemplateVariableDefinition)
        self.assertEqual(template_var.variable_name, "test_int_var")
        self.assertEqual(template_var.description, "Test integer variable")
        self.assertEqual(template_var.default, 42)

    def test_number_var_instantiable_with_float_and_converts_to_float_vardef(self) -> None:
        # Arrange
        tf_var = TerraformNumberVariableDefinition(
            variable_name="test_float_var",
            description="Test float variable",
            default=3.14,
        )

        # Act
        template_var = tf_var.to_template_definition()

        # Assert
        self.assertIsInstance(template_var, FloatTemplateVariableDefinition)
        self.assertEqual(template_var.variable_name, "test_float_var")
        self.assertEqual(template_var.description, "Test float variable")
        self.assertEqual(template_var.default, 3.14)

    def test_number_var_instantiable_and_converts_to_anynumeric_vardef(self) -> None:
        # Arrange
        tf_var = TerraformNumberVariableDefinition(
            variable_name="test_numeric_var",
            description="Test numeric variable",
            # No default provided
        )

        # Act
        template_var = tf_var.to_template_definition()

        # Assert
        self.assertIsInstance(template_var, AnyNumericTemplateVariableDefinition)
        self.assertEqual(template_var.variable_name, "test_numeric_var")
        self.assertEqual(template_var.description, "Test numeric variable")
        self.assertIsNone(template_var.default)

    def test_bool_var_instantiable_and_converts_to_bool_vardef(self) -> None:
        # Arrange
        tf_var = TerraformBoolVariableDefinition(
            variable_name="test_bool_var",
            description="Test boolean variable",
            default=True,
        )

        # Act
        template_var = tf_var.to_template_definition()

        # Assert
        self.assertIsInstance(template_var, BoolTemplateVariableDefinition)
        self.assertEqual(template_var.variable_name, "test_bool_var")
        self.assertEqual(template_var.description, "Test boolean variable")
        self.assertEqual(template_var.default, True)

    @parameterized.expand(
        [
            (
                "string_type",
                {"variable_name": "str_var", "tf_type": TerraformType.STRING},
                TerraformStrVariableDefinition,
            ),
            (
                "number_type",
                {"variable_name": "num_var", "tf_type": TerraformType.NUMBER},
                TerraformNumberVariableDefinition,
            ),
            (
                "bool_type",
                {"variable_name": "bool_var", "tf_type": TerraformType.BOOL},
                TerraformBoolVariableDefinition,
            ),
            (
                "list_str_type",
                {"variable_name": "list_var", "tf_type": TerraformType.LIST_STR},
                TerraformListOfStrVariableDefinition,
            ),
            (
                "map_str_type",
                {"variable_name": "map_var", "tf_type": TerraformType.MAP_STR},
                TerraformMapOfStrVariableDefinition,
            ),
            (
                "list_map_str_type",
                {"variable_name": "list_map_var", "tf_type": TerraformType.LIST_MAP_STR},
                TerraformListOfMapStrVariableDefinition,
            ),
        ]
    )
    def test_create_tf_variable_definition_maps_to_the_right_class(
        self, _name: str, config: dict, expected_class: type
    ) -> None:
        # Act
        result = create_tf_variable_definition(config)

        # Assert
        self.assertIsInstance(result, expected_class)
        self.assertEqual(result.variable_name, config["variable_name"])

    @patch("jupyter_deploy.engine.vardefs.StrTemplateVariableDefinition", spec=True)
    def test_to_tf_var_option_wraps_empty_str(self, mock_var_def: Mock) -> None:
        # Arrange
        mock_var_def.variable_name = "var1"
        mock_var_def.assigned_value = ""

        # Act
        result = to_tf_var_option(mock_var_def)

        # Assert
        self.assertEqual(result, ["-var", 'var1=""'])

    @patch("jupyter_deploy.engine.vardefs.StrTemplateVariableDefinition", spec=True)
    def test_to_tf_var_option_converts_none(self, mock_var_def: Mock) -> None:
        # Arrange
        mock_var_def.variable_name = "var1"
        mock_var_def.assigned_value = None

        # Act
        result = to_tf_var_option(mock_var_def)

        # Assert
        self.assertEqual(result, ["-var", "var1=null"])

    @parameterized.expand(
        [
            ("string_value", "hello", StrTemplateVariableDefinition, ["-var", "string_value=hello"]),
            ("int_value", 42, IntTemplateVariableDefinition, ["-var", "int_value=42"]),
            ("float_value", 3.14, FloatTemplateVariableDefinition, ["-var", "float_value=3.14"]),
            ("bool_true", True, BoolTemplateVariableDefinition, ["-var", "bool_true=true"]),
            ("bool_false", False, BoolTemplateVariableDefinition, ["-var", "bool_false=false"]),
            ("list_value", ["a", "b", "c"], ListStrTemplateVariableDefinition, ["-var", 'list_value=["a", "b", "c"]']),
            (
                "dict_value",
                {"key1": "val1", "key2": "val2"},
                DictStrTemplateVariableDefinition,
                ["-var", 'dict_value={"key1": "val1", "key2": "val2"}'],
            ),
            (
                "list_dict_str_value",
                [{"k1": "v1", "k2": "v2"}, {"k3": "v3"}],
                ListMapStrTemplateVariableDefinition,
                ["-var", 'list_dict_str_value=[{"k1": "v1", "k2": "v2"}, {"k3": "v3"}]'],
            ),
        ]
    )
    def test_to_tf_var_option(self, name: str, input_value: Any, clz: type, expected_output: list[str]) -> None:
        # Arrange
        mock_var_def = Mock(spec=clz)
        mock_var_def.variable_name = name
        mock_var_def.assigned_value = input_value

        # Act
        result = to_tf_var_option(mock_var_def)

        # Assert
        self.assertEqual(result, expected_output)



from jupyter_deploy.engine.engine_down import EngineDownHandler
from jupyter_deploy.engine.enum import EngineType
from jupyter_deploy.engine.terraform import tf_down
from jupyter_deploy.handlers.base_project_handler import BaseProjectHandler


class DownHandler(BaseProjectHandler):
    _handler: EngineDownHandler

    def __init__(self) -> None:
        """Base class to manage the down command of a jupyter-deploy project."""
        super().__init__()

        if self.engine == EngineType.TERRAFORM:
            self._handler = tf_down.TerraformDownHandler(
                project_path=self.project_path, project_manifest=self.project_manifest
            )
        else:
            raise NotImplementedError(f"DownHandler implementation not found for engine: {self.engine}")

    def destroy(self, auto_approve: bool = False) -> None:
        """Destroy the infrastructure resources.

        Args:
            auto_approve: Whether to auto-approve the destruction without prompting.
        """
        return self._handler.destroy(auto_approve)



from jupyter_deploy.engine.engine_outputs import EngineOutputsHandler
from jupyter_deploy.engine.enum import EngineType
from jupyter_deploy.engine.terraform import tf_outputs, tf_variables
from jupyter_deploy.handlers.base_project_handler import BaseProjectHandler
from jupyter_deploy.provider import manifest_command_runner as cmd_runner
from jupyter_deploy.provider.resolved_clidefs import StrResolvedCliParameter


class ServerHandler(BaseProjectHandler):
    """Handler class to directly interact with a jupyter server app."""

    _output_handler: EngineOutputsHandler

    def __init__(self) -> None:
        """Instantiate the Users handler."""
        super().__init__()

        if self.engine == EngineType.TERRAFORM:
            self._output_handler = tf_outputs.TerraformOutputsHandler(
                project_path=self.project_path, project_manifest=self.project_manifest
            )
            self._variable_handler = tf_variables.TerraformVariablesHandler(
                project_path=self.project_path, project_manifest=self.project_manifest
            )
        else:
            raise NotImplementedError(f"OutputsHandler implementation not found for engine: {self.engine}")

    def get_server_status(self) -> str:
        """Sends an health check to the jupyter server app, return status."""
        command = self.project_manifest.get_command("server.status")
        console = self.get_console()
        runner = cmd_runner.ManifestCommandRunner(
            console=console, output_handler=self._output_handler, variable_handler=self._variable_handler
        )
        runner.run_command_sequence(command, cli_paramdefs={})
        return runner.get_result_value(command, "server.status", str)

    def start_server(self, service: str) -> None:
        """Start the jupyter server and optionally all its sidecars."""
        command = self.project_manifest.get_command("server.start")
        validated_service = self.project_manifest.get_validated_service(service, allow_all=True)
        console = self.get_console()
        runner = cmd_runner.ManifestCommandRunner(
            console=console, output_handler=self._output_handler, variable_handler=self._variable_handler
        )
        runner.run_command_sequence(
            command,
            cli_paramdefs={
                "action": StrResolvedCliParameter(parameter_name="action", value="start"),
                "service": StrResolvedCliParameter(parameter_name="service", value=validated_service),
            },
        )

    def stop_server(self, service: str) -> None:
        """Stop the jupyter server and optionally all its sidecars."""
        command = self.project_manifest.get_command("server.stop")
        validated_service = self.project_manifest.get_validated_service(service, allow_all=True)
        console = self.get_console()
        runner = cmd_runner.ManifestCommandRunner(
            console=console, output_handler=self._output_handler, variable_handler=self._variable_handler
        )
        runner.run_command_sequence(
            command,
            cli_paramdefs={
                "action": StrResolvedCliParameter(parameter_name="action", value="stop"),
                "service": StrResolvedCliParameter(parameter_name="service", value=validated_service),
            },
        )

    def restart_server(self, service: str) -> None:
        """Restart the jupyter-server and optionally all its sidecars."""
        command = self.project_manifest.get_command("server.restart")
        validated_service = self.project_manifest.get_validated_service(service, allow_all=True)
        console = self.get_console()
        runner = cmd_runner.ManifestCommandRunner(
            console=console, output_handler=self._output_handler, variable_handler=self._variable_handler
        )
        runner.run_command_sequence(
            command,
            cli_paramdefs={
                "action": StrResolvedCliParameter(parameter_name="action", value="restart"),
                "service": StrResolvedCliParameter(parameter_name="service", value=validated_service),
            },
        )

    def get_server_logs(self, service: str, extra: list[str]) -> tuple[str, str]:
        """Sends a logs command to the host for the service, return the stdout, stderr."""
        command = self.project_manifest.get_command("server.logs")
        validated_service = self.project_manifest.get_validated_service(service, allow_all=False)
        console = self.get_console()
        runner = cmd_runner.ManifestCommandRunner(
            console=console, output_handler=self._output_handler, variable_handler=self._variable_handler
        )

        runner.run_command_sequence(
            command,
            cli_paramdefs={
                "service": StrResolvedCliParameter(parameter_name="service", value=validated_service),
                "extra": StrResolvedCliParameter(parameter_name="extra", value=" ".join(extra)),
            },
        )
        stdout = runner.get_result_value(command, "server.logs", str)
        stderr = runner.get_result_value(command, "server.errors", str)
        return stdout, stderr



import unittest
from pathlib import Path
from typing import Any

import yaml

from jupyter_deploy.engine.enum import EngineType
from jupyter_deploy.manifest import InvalidServiceError, JupyterDeployManifestV1


class TestJupyterDeployManifestV1(unittest.TestCase):
    manifest_v1_content: str
    manifest_v1_parsed_content: Any

    @classmethod
    def setUpClass(cls) -> None:
        mock_manifest_path = Path(__file__).parent / "mock_manifest.yaml"
        with open(mock_manifest_path) as f:
            cls.manifest_v1_content = f.read()
        cls.manifest_v1_parsed_content = yaml.safe_load(cls.manifest_v1_content)

    def test_can_parse_manifest_v1(self) -> None:
        JupyterDeployManifestV1(
            **self.manifest_v1_parsed_content  # type: ignore
        )

    def test_manifest_v1_get_engine(self) -> None:
        manifest = JupyterDeployManifestV1(
            **self.manifest_v1_parsed_content  # type: ignore
        )
        self.assertEqual(manifest.get_engine(), EngineType.TERRAFORM)

    def test_manifest_v1_get_declared_value_happy_path(self) -> None:
        manifest = JupyterDeployManifestV1(
            **self.manifest_v1_parsed_content  # type: ignore
        )
        self.assertEqual(
            manifest.get_declared_value("aws_region"),
            manifest.values[1],  # type: ignore
        )

    def test_manifest_v1_get_declared_value_raises_not_implement_error(self) -> None:
        manifest = JupyterDeployManifestV1(
            **self.manifest_v1_parsed_content  # type: ignore
        )
        with self.assertRaises(NotImplementedError):
            manifest.get_declared_value("i_am_not_declared")

    def test_manifest_v1_get_command_happy_path(self) -> None:
        manifest = JupyterDeployManifestV1(
            **self.manifest_v1_parsed_content  # type: ignore
        )
        manifest.get_command("server.status")  # should not raise

    def test_manifest_v1_not_found_command_raises_not_implemented_error(self) -> None:
        manifest = JupyterDeployManifestV1(
            **self.manifest_v1_parsed_content  # type: ignore
        )
        with self.assertRaises(NotImplementedError):
            manifest.get_command("cmd_does_not_exist")

    def test_manifest_v1_has_command_found(self) -> None:
        manifest = JupyterDeployManifestV1(
            **self.manifest_v1_parsed_content  # type: ignore
        )
        self.assertTrue(manifest.has_command("host.status"))

    def test_manifest_v1_has_command_not_found(self) -> None:
        manifest = JupyterDeployManifestV1(
            **self.manifest_v1_parsed_content  # type: ignore
        )
        self.assertFalse(manifest.has_command("i.do.not.exist"))

    def test_manifest_v1_get_services(self) -> None:
        manifest = JupyterDeployManifestV1(
            **self.manifest_v1_parsed_content  # type: ignore
        )
        self.assertListEqual(manifest.get_services(), ["jupyter", "traefik", "oauth"])

    def test_manifest_v1_get_validated_service(self) -> None:
        manifest = JupyterDeployManifestV1(
            **self.manifest_v1_parsed_content  # type: ignore
        )
        for svc in ["jupyter", "traefik", "oauth"]:
            self.assertEqual(manifest.get_validated_service(svc), svc)

    def test_manifest_v1_get_validated_service_default_return_first_value(self) -> None:
        manifest = JupyterDeployManifestV1(
            **self.manifest_v1_parsed_content  # type: ignore
        )
        self.assertEqual(manifest.get_validated_service("default"), "jupyter")

    def test_manifest_v1_get_validated_service_all_allowed(self) -> None:
        manifest = JupyterDeployManifestV1(
            **self.manifest_v1_parsed_content  # type: ignore
        )
        self.assertEqual(manifest.get_validated_service("all", allow_all=True), "all")

    def test_manifest_v1_get_validated_service_all_disallowed(self) -> None:
        manifest = JupyterDeployManifestV1(
            **self.manifest_v1_parsed_content  # type: ignore
        )
        with self.assertRaises(InvalidServiceError):
            manifest.get_validated_service("all", allow_all=False)



"""Terraform implementation of the `up` handler."""

from pathlib import Path

from rich import console as rich_console

from jupyter_deploy import cmd_utils
from jupyter_deploy.engine.engine_up import EngineUpHandler
from jupyter_deploy.engine.enum import EngineType
from jupyter_deploy.engine.terraform.tf_constants import (
    TF_APPLY_CMD,
    TF_AUTO_APPROVE_CMD_OPTION,
    TF_DEFAULT_PLAN_FILENAME,
    TF_ENGINE_DIR,
)


class TerraformUpHandler(EngineUpHandler):
    """Up handler implementation for terraform projects."""

    def __init__(self, project_path: Path) -> None:
        self.engine_dir_path = project_path / TF_ENGINE_DIR
        super().__init__(project_path=project_path, engine=EngineType.TERRAFORM, engine_dir_path=self.engine_dir_path)

    def get_default_config_filename(self) -> str:
        return TF_DEFAULT_PLAN_FILENAME

    def apply(self, config_file_path: Path, auto_approve: bool = False) -> None:
        console = rich_console.Console()

        apply_cmd = TF_APPLY_CMD.copy()
        if auto_approve:
            apply_cmd.append(TF_AUTO_APPROVE_CMD_OPTION)
        apply_cmd.append(str(config_file_path.absolute()))

        retcode, timed_out = cmd_utils.run_cmd_and_pipe_to_terminal(apply_cmd, exec_dir=self.engine_dir_path)

        if retcode != 0 or timed_out:
            console.print(":x: Error applying Terraform plan.", style="red")
            return

        console.print("Infrastructure changes applied successfully.", style="green")



from enum import Enum


class TerraformType(str, Enum):
    """The 'type' attribute in the variables.tf file as surfaced by python-hcl2."""

    STRING = "string"
    NUMBER = "number"
    BOOL = "bool"
    LIST_STR = "${list(string)}"
    MAP_STR = "${map(string)}"
    LIST_MAP_STR = "${list(map(string))}"



# mypy: disable-error-code=attr-defined

import json
import unittest
from pathlib import Path

from pydantic import ValidationError

from jupyter_deploy.engine.terraform.tf_plan import (
    TerraformPlan,
    TerraformPlanVariableContent,
    extract_variables_from_json_plan,
    format_plan_variables,
    format_terraform_value,
    format_values_for_dot_tfvars,
)


class TestFormatTerraformValue(unittest.TestCase):
    def test_null_value(self) -> None:
        self.assertEqual(format_terraform_value(None), "null")

    def test_str_value(self) -> None:
        self.assertEqual(format_terraform_value("hello"), '"hello"')

    def test_empty_str_value(self) -> None:
        self.assertEqual(format_terraform_value(""), '""')

    def test_bool_true_value(self) -> None:
        self.assertEqual(format_terraform_value(True), "true")

    def test_bool_false_value(self) -> None:
        self.assertEqual(format_terraform_value(False), "false")

    def test_list_str_value(self) -> None:
        self.assertEqual(format_terraform_value(["a", "b"]), '[\n"a",\n"b",\n]')

    def test_list_int_value(self) -> None:
        self.assertEqual(format_terraform_value([1, 2]), "[\n1,\n2,\n]")

    def test_list_float_value(self) -> None:
        self.assertEqual(format_terraform_value([1.1, 2.2]), "[\n1.1,\n2.2,\n]")

    def test_empty_list_value(self) -> None:
        self.assertEqual(format_terraform_value([]), "[]")

    def test_dict_str_str_value(self) -> None:
        result = format_terraform_value({"key": "value"})
        self.assertEqual(result, '{\nkey = "value"\n}')

    def test_dict_str_int_value(self) -> None:
        result = format_terraform_value({"key": 123})
        self.assertEqual(result, "{\nkey = 123\n}")

    def test_dict_str_float_value(self) -> None:
        result = format_terraform_value({"key": 1.23})
        self.assertEqual(result, "{\nkey = 1.23\n}")

    def test_empty_dict_value(self) -> None:
        self.assertEqual(format_terraform_value({}), "{}")


class TestExtractVariablesFromPlan(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        mock_plan_path = Path(__file__).parent / "mock_plan.json"
        with open(mock_plan_path) as f:
            cls.plan_content = f.read()
        cls.plan = TerraformPlan(**json.loads(cls.plan_content))

    def test_happy_case(self) -> None:
        cls = self.__class__
        result = extract_variables_from_json_plan(cls.plan_content)

        expect_vars = {k: v for k, v in cls.plan.variables.items() if "secret" not in k}
        expect_secrets = {k: v for k, v in cls.plan.variables.items() if "secret" in k}

        self.assertTupleEqual(result, (expect_vars, expect_secrets))

    def test_invalid_json_raise_value_error(self) -> None:
        cls = self.__class__

        with self.assertRaises(ValueError):
            extract_variables_from_json_plan(cls.plan_content[:-2])

    def test_non_dict_json_raise_value_error(self) -> None:
        with self.assertRaises(ValueError):
            extract_variables_from_json_plan(json.dumps(["I should be a dict"]))

    def test_no_variables_key_raise_pydantic_validation_error(self) -> None:
        cls = self.__class__
        no_variables_plan = json.loads(cls.plan_content)
        del no_variables_plan["variables"]

        with self.assertRaises(ValidationError):
            extract_variables_from_json_plan(json.dumps(no_variables_plan))

    def test_no_configuration_key_raise_pydantic_validation_error(self) -> None:
        cls = self.__class__
        no_config_plan = json.loads(cls.plan_content)
        del no_config_plan["configuration"]

        with self.assertRaises(ValidationError):
            extract_variables_from_json_plan(json.dumps(no_config_plan))

    def test_no_config_root_variables_key_raise_pydantic_validation_error(self) -> None:
        cls = self.__class__
        modified_plan = json.loads(cls.plan_content)
        del modified_plan["configuration"]["root_module"]["variables"]

        with self.assertRaises(ValidationError):
            extract_variables_from_json_plan(json.dumps(modified_plan))


class TestFormatPlanVariables(unittest.TestCase):
    def test_happy_case(self) -> None:
        vars = {
            "var_str": TerraformPlanVariableContent(value="value1"),
            "var_int": TerraformPlanVariableContent(value=123),
            "var_float": TerraformPlanVariableContent(value=3.1459),
            "var_bool": TerraformPlanVariableContent(value=True),
            "var_null": TerraformPlanVariableContent(value=None),
            "var_empty_dict": TerraformPlanVariableContent(value={}),
            "var_dict": TerraformPlanVariableContent(value={"key1": "val1", "key2": "val2"}),
            "var_dict_int": TerraformPlanVariableContent(value={"key1": 10, "key2": 11}),
            "var_empty_list": TerraformPlanVariableContent(value=[]),
            "var_list": TerraformPlanVariableContent(value=["a", "b"]),
        }
        result = format_plan_variables(vars)
        self.assertGreaterEqual(len(result), len(vars.keys()))  # allow for top-level comments
        self.assertIn('var_str = "value1"\n', result)
        self.assertIn("var_int = 123\n", result)
        self.assertIn("var_float = 3.1459\n", result)
        self.assertIn("var_bool = true\n", result)
        self.assertIn("var_null = null\n", result)
        self.assertIn("var_empty_dict = {}\n", result)
        self.assertIn('var_dict = {\nkey1 = "val1"\nkey2 = "val2"\n}\n', result)
        self.assertIn("var_dict_int = {\nkey1 = 10\nkey2 = 11\n}\n", result)
        self.assertIn("var_empty_list = []\n", result)
        self.assertIn('var_list = [\n"a",\n"b",\n]\n', result)

    def test_empty_vars_return_empty_list(self) -> None:
        vars: dict[str, TerraformPlanVariableContent] = {}
        result = format_plan_variables(vars)
        self.assertEqual(result, [])


class TestFormatValuesForDotTfvars(unittest.TestCase):
    def test_happy_case(self) -> None:
        vars = {
            "var_str": "value1",
            "var_int": 123,
            "var_float": 3.1459,
            "var_bool": True,
            "var_null": None,
            "var_empty_dict": {},
            "var_dict": {"key1": "val1", "key2": "val2"},
            "var_dict_int": {"key1": 10, "key2": 11},
            "var_empty_list": [],
            "var_list": ["a", "b"],
        }
        result = format_values_for_dot_tfvars(vars)
        self.assertGreaterEqual(len(result), len(vars.keys()))  # allow for include comments
        self.assertIn('var_str = "value1"\n', result)
        self.assertIn("var_int = 123\n", result)
        self.assertIn("var_float = 3.1459\n", result)
        self.assertIn("var_bool = true\n", result)
        self.assertIn("var_null = null\n", result)
        self.assertIn("var_empty_dict = {}\n", result)
        self.assertIn('var_dict = {\nkey1 = "val1"\nkey2 = "val2"\n}\n', result)
        self.assertIn("var_dict_int = {\nkey1 = 10\nkey2 = 11\n}\n", result)
        self.assertIn("var_empty_list = []\n", result)
        self.assertIn('var_list = [\n"a",\n"b",\n]\n', result)

    def test_empty_vars_return_empty_list(self) -> None:
        vars: dict[str, TerraformPlanVariableContent] = {}
        result = format_values_for_dot_tfvars(vars)
        self.assertEqual(result, [])



import sys
import unittest
from collections.abc import Generator
from contextlib import contextmanager
from unittest.mock import Mock, patch

from jupyter_deploy.engine.engine_outputs import EngineOutputsHandler
from jupyter_deploy.engine.outdefs import StrTemplateOutputDefinition
from jupyter_deploy.provider.enum import ProviderType
from jupyter_deploy.provider.instruction_runner import InstructionRunner


class TestInstructionRunnerFactory(unittest.TestCase):
    def setUp(self) -> None:
        # api providers modules mocks
        self.mock_aws_api_runner = Mock(spec=InstructionRunner)
        self.mock_aws_api_runner_cls = Mock()
        self.mock_aws_api_runner_cls.return_value = self.mock_aws_api_runner

        self.mock_aws_runner_module = Mock()
        self.mock_aws_runner_module.AwsApiRunner = self.mock_aws_api_runner_cls

        # outputs handler mock
        self.mock_outputs_handler = Mock(spec=EngineOutputsHandler)
        self.mock_str_template_output_def = Mock(spec=StrTemplateOutputDefinition)
        self.mock_str_template_output_def.value = "us-west-2"
        self.mock_get_declared_output_def = Mock()
        self.mock_get_declared_output_def.return_value = self.mock_str_template_output_def
        self.mock_outputs_handler.get_declared_output_def = self.mock_get_declared_output_def

    @contextmanager
    def patch_provider_runner_modules(self) -> Generator:
        """Patch all provider runner modules from the correct import location.

        This is subtle:
            - when running the global `uv run pytest`, the provider modules are already imported
            - when running `uv run pytest /path/to/this/test_module.py, they are not (runtime import)

        This results in different behavior where test may pass when run in isolation, and fail
        when run globally - obviously we do not want that.
        """

        # case 1: module(s) already imported - global `uv run pytest` call
        if "jupyter_deploy.provider.aws.aws_runner" not in sys.modules:
            with patch.dict(sys.modules, {"jupyter_deploy.provider.aws.aws_runner": self.mock_aws_runner_module}):
                yield
        # case 2: module(s) not imported yet - `uv run pytest /path/to/this/test_module.py`
        else:
            with patch("jupyter_deploy.provider.aws.aws_runner", self.mock_aws_runner_module):
                yield

    def test_does_not_create_any_runner_provider_on_class_setup(self) -> None:
        with self.patch_provider_runner_modules():
            from jupyter_deploy.provider.instruction_runner_factory import InstructionRunnerFactory

            InstructionRunnerFactory._provider_runner_map = {}

            self.mock_aws_api_runner_cls.assert_not_called()

            # Verify that the provider runner map is empty
            self.assertEqual({}, InstructionRunnerFactory._provider_runner_map)

    def test_imports_aws_provider_at_runtime_only_and_return_it(self) -> None:
        # Execute
        with self.patch_provider_runner_modules():
            from jupyter_deploy.provider.instruction_runner_factory import InstructionRunnerFactory

            InstructionRunnerFactory._provider_runner_map = {}

            runner = InstructionRunnerFactory.get_provider_instruction_runner("aws", self.mock_outputs_handler)

            # Verify
            self.mock_get_declared_output_def.assert_called_once_with("aws_region", StrTemplateOutputDefinition)

            self.assertEqual(self.mock_aws_api_runner, runner)
            self.assertEqual(
                {ProviderType.AWS: self.mock_aws_api_runner}, InstructionRunnerFactory._provider_runner_map
            )
            self.mock_aws_api_runner_cls.assert_called_once_with(region_name="us-west-2")

    def test_aws_provider_raises_if_output_provider_cannot_get_the_region(self) -> None:
        # Setup
        mock_outputs_handler = Mock(spec=EngineOutputsHandler)
        mock_outputs_handler.get_declared_output_def.side_effect = ValueError("Region not found")

        # Execute and verify
        with self.assertRaises(ValueError) as context, self.patch_provider_runner_modules():
            from jupyter_deploy.provider.instruction_runner_factory import InstructionRunnerFactory

            InstructionRunnerFactory._provider_runner_map = {}

            InstructionRunnerFactory.get_provider_instruction_runner("aws", mock_outputs_handler)

            self.assertEqual("Region not found", str(context.exception))
            self.mock_get_declared_output_def.assert_called_once_with("aws_region", StrTemplateOutputDefinition)
            self.mock_aws_api_runner_cls.assert_not_called()

    def test_recycle_aws_runner_provider_for_same_output_handler(self) -> None:
        with self.patch_provider_runner_modules():
            from jupyter_deploy.provider.instruction_runner_factory import InstructionRunnerFactory

            InstructionRunnerFactory._provider_runner_map = {}

            # First call
            first_result = InstructionRunnerFactory.get_provider_instruction_runner("aws", self.mock_outputs_handler)

            # Second call with same output handler
            second_result = InstructionRunnerFactory.get_provider_instruction_runner("aws", self.mock_outputs_handler)

            # Verify
            self.assertEqual(first_result, second_result)
            self.assertEqual(
                {ProviderType.AWS: self.mock_aws_api_runner}, InstructionRunnerFactory._provider_runner_map
            )
            self.mock_aws_api_runner_cls.assert_called_once()

    def test_recycle_aws_runner_provider_for_different_output_handler(self) -> None:
        # Setup
        mock_outputs_handler2 = Mock(spec=EngineOutputsHandler)
        mock_str_template_output_def2 = Mock(spec=StrTemplateOutputDefinition)
        mock_str_template_output_def2.value = "us-west-2"  # Same region
        mock_outputs_handler2.get_declared_output_def.return_value = mock_str_template_output_def2

        with self.patch_provider_runner_modules():
            from jupyter_deploy.provider.instruction_runner_factory import InstructionRunnerFactory

            InstructionRunnerFactory._provider_runner_map = {}

            # First call
            first_result = InstructionRunnerFactory.get_provider_instruction_runner("aws", self.mock_outputs_handler)

            # Second call with same output handler
            second_result = InstructionRunnerFactory.get_provider_instruction_runner("aws", mock_outputs_handler2)

            # Verify
            self.assertEqual(first_result, second_result)
            self.assertEqual(
                {ProviderType.AWS: self.mock_aws_api_runner}, InstructionRunnerFactory._provider_runner_map
            )
            self.mock_aws_api_runner_cls.assert_called_once()

    def test_raise_not_value_error_on_unmatched_provider(self) -> None:
        mock_outputs_handler = Mock(spec=EngineOutputsHandler)

        with (
            self.patch_provider_runner_modules(),
            self.assertRaises(ValueError),
        ):
            from jupyter_deploy.provider.instruction_runner_factory import InstructionRunnerFactory

            InstructionRunnerFactory._provider_runner_map = {}

            InstructionRunnerFactory.get_provider_instruction_runner("onpremises", mock_outputs_handler)
            self.mock_aws_api_runner_cls.assert_not_called()
            self.mock_get_declared_output_def.assert_not_called()



import unittest
from pathlib import Path
from unittest.mock import Mock, patch

import yaml

from jupyter_deploy.engine.enum import EngineType
from jupyter_deploy.handlers.access.organization_handler import OrganizationHandler
from jupyter_deploy.manifest import JupyterDeployManifest, JupyterDeployManifestV1
from jupyter_deploy.provider.resolved_clidefs import StrResolvedCliParameter


class TestOrganizationHandler(unittest.TestCase):
    mock_full_manifest: JupyterDeployManifest

    @classmethod
    def setUpClass(cls) -> None:
        full_manifest_path = Path(__file__).parent.parent.parent / "mock_manifest.yaml"
        with open(full_manifest_path) as f:
            manifest_content = f.read()
        manifest_parsed_content = yaml.safe_load(manifest_content)
        cls.mock_full_manifest = JupyterDeployManifestV1(
            **manifest_parsed_content  # type: ignore
        )

    def get_mock_manifest_and_fns(self) -> tuple[Mock, dict[str, Mock]]:
        """Return mock manifest with functions defined as mock."""
        mock_manifest = Mock()
        mock_get_engine = Mock()
        mock_get_command = Mock()
        mock_get_engine.return_value = EngineType.TERRAFORM
        mock_get_command.return_value = Mock()
        mock_manifest.get_command = mock_get_command
        mock_manifest.get_engine = mock_get_engine
        return mock_manifest, {"get_command": mock_get_command, "get_engine": mock_get_engine}

    def get_mock_outputs_handler_and_fns(self) -> tuple[Mock, dict[str, Mock]]:
        """Return mock output handler with functions defined as mock."""
        mock_output_handler = Mock()
        return mock_output_handler, {}

    def get_mock_manifest_cmd_runner_and_fns(self) -> tuple[Mock, dict[str, Mock]]:
        """Return mock manifest cmd runner with functions defined as mock."""
        mock_cmd_runner_handler = Mock()
        mock_run_command_sequence = Mock()
        mock_get_result_value = Mock()
        mock_update_variables = Mock()

        mock_cmd_runner_handler.run_command_sequence = mock_run_command_sequence
        mock_cmd_runner_handler.get_result_value = mock_get_result_value
        mock_cmd_runner_handler.update_variables = mock_update_variables

        mock_run_command_sequence.return_value = (True, {})
        mock_get_result_value.return_value = "org-name"

        return mock_cmd_runner_handler, {
            "run_command_sequence": mock_run_command_sequence,
            "get_result_value": mock_get_result_value,
            "update_variables": mock_update_variables,
        }

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("jupyter_deploy.engine.terraform.tf_outputs.TerraformOutputsHandler")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    @patch("pathlib.Path.cwd")
    def test_can_instantiate_terraform_project(
        self,
        mock_cwd: Mock,
        mock_tf_variables_handler: Mock,
        mock_tf_outputs_handler: Mock,
        mock_retrieve_manifest: Mock,
    ) -> None:
        path = Path("/some/cur/dir")
        mock_cwd.return_value = path

        mock_manifest, _ = self.get_mock_manifest_and_fns()
        mock_retrieve_manifest.return_value = mock_manifest

        mock_output_handler = self.get_mock_outputs_handler_and_fns()[0]
        mock_tf_outputs_handler.return_value = mock_output_handler

        mock_variable_handler = Mock()
        mock_tf_variables_handler.return_value = mock_variable_handler

        handler = OrganizationHandler()

        mock_retrieve_manifest.assert_called_once()
        mock_tf_outputs_handler.assert_called_once_with(project_path=path, project_manifest=mock_manifest)
        mock_tf_variables_handler.assert_called_once_with(project_path=path, project_manifest=mock_manifest)
        self.assertEqual(handler._output_handler, mock_output_handler)
        self.assertEqual(handler._variable_handler, mock_variable_handler)
        self.assertEqual(handler.engine, EngineType.TERRAFORM)

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("jupyter_deploy.engine.terraform.tf_outputs.TerraformOutputsHandler")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    def test_organization_methods_raise_not_implemented_error_if_manifest_does_not_define_cmd(
        self, mock_tf_variables_handler: Mock, mock_tf_outputs_handler: Mock, mock_retrieve_manifest: Mock
    ) -> None:
        mock_tf_outputs_handler.return_value = self.get_mock_outputs_handler_and_fns()[0]
        mock_tf_variables_handler.return_value = Mock()

        # Create a manifest with no commands defined
        no_cmd_manifest = JupyterDeployManifestV1(
            **{  # type: ignore
                "schema_version": 1,
                "template": {
                    "name": "mock-template-name",
                    "engine": "terraform",
                    "version": "1.0.0",
                },
            }
        )
        mock_retrieve_manifest.return_value = no_cmd_manifest
        handler = OrganizationHandler()

        with self.assertRaises(NotImplementedError):
            handler.set_organization("org-name")

        with self.assertRaises(NotImplementedError):
            handler.unset_organization()

        with self.assertRaises(NotImplementedError):
            handler.get_organization()

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("jupyter_deploy.engine.terraform.tf_outputs.TerraformOutputsHandler")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    @patch("jupyter_deploy.provider.manifest_command_runner.ManifestCommandRunner")
    def test_organization_methods_run_against_actual_manifest(
        self,
        mock_cmd_runner_class: Mock,
        mock_tf_variables_handler: Mock,
        mock_tf_outputs_handler: Mock,
        mock_retrieve_manifest: Mock,
    ) -> None:
        mock_retrieve_manifest.return_value = self.mock_full_manifest
        mock_cmd_runner_class.return_value = self.get_mock_manifest_cmd_runner_and_fns()[0]
        mock_tf_outputs_handler.return_value = self.get_mock_outputs_handler_and_fns()[0]
        mock_tf_variables_handler.return_value = Mock()

        handler = OrganizationHandler()

        # verify methods work
        handler.set_organization("org-name")
        handler.unset_organization()

        org_name = handler.get_organization()
        self.assertEqual(org_name, "org-name")

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("jupyter_deploy.engine.terraform.tf_outputs.TerraformOutputsHandler")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    @patch("jupyter_deploy.provider.manifest_command_runner.ManifestCommandRunner")
    def test_organization_methods_raises_if_run_command_raises(
        self,
        mock_cmd_runner_class: Mock,
        mock_tf_variables_handler: Mock,
        mock_tf_outputs_handler: Mock,
        mock_retrieve_manifest: Mock,
    ) -> None:
        # Setup
        mock_manifest, _ = self.get_mock_manifest_and_fns()
        mock_retrieve_manifest.return_value = mock_manifest

        mock_tf_outputs_handler.return_value = self.get_mock_outputs_handler_and_fns()[0]
        mock_tf_variables_handler.return_value = Mock()

        mock_cmd_runner, mock_cmd_runner_fns = self.get_mock_manifest_cmd_runner_and_fns()
        mock_cmd_runner_class.return_value = mock_cmd_runner
        mock_cmd_runner_fns["run_command_sequence"].side_effect = RuntimeError()

        handler = OrganizationHandler()

        # verify methods raise
        with self.assertRaises(RuntimeError):
            handler.set_organization("org-name")
        with self.assertRaises(RuntimeError):
            handler.unset_organization()
        with self.assertRaises(RuntimeError):
            handler.get_organization()

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("jupyter_deploy.engine.terraform.tf_outputs.TerraformOutputsHandler")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    @patch("jupyter_deploy.provider.manifest_command_runner.ManifestCommandRunner")
    def test_organization_methods_raise_if_get_command_result_raises(
        self,
        mock_cmd_runner_class: Mock,
        mock_tf_variables_handler: Mock,
        mock_tf_outputs_handler: Mock,
        mock_retrieve_manifest: Mock,
    ) -> None:
        # Setup
        mock_manifest, _ = self.get_mock_manifest_and_fns()
        mock_retrieve_manifest.return_value = mock_manifest

        mock_tf_outputs_handler.return_value = self.get_mock_outputs_handler_and_fns()[0]
        mock_tf_variables_handler.return_value = Mock()

        mock_cmd_runner, mock_cmd_runner_fns = self.get_mock_manifest_cmd_runner_and_fns()
        mock_cmd_runner_class.return_value = mock_cmd_runner
        mock_cmd_runner_fns["get_result_value"].side_effect = KeyError()

        handler = OrganizationHandler()

        # verify method raises
        # add only commands that return a result here
        with self.assertRaises(KeyError):
            handler.get_organization()

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    @patch("jupyter_deploy.engine.terraform.tf_outputs.TerraformOutputsHandler")
    @patch("jupyter_deploy.provider.manifest_command_runner.ManifestCommandRunner")
    @patch("rich.console.Console")
    def test_set_organization_calls_run_command_sequence_with_correct_params(
        self,
        mock_console_class: Mock,
        mock_cmd_runner_class: Mock,
        mock_tf_outputs_handler: Mock,
        mock_tf_variables_handler: Mock,
        mock_retrieve_manifest: Mock,
    ) -> None:
        # Setup
        mock_cmd = Mock()
        mock_manifest, mock_manifest_fns = self.get_mock_manifest_and_fns()
        mock_manifest_fns["get_command"].return_value = mock_cmd
        mock_retrieve_manifest.return_value = mock_manifest

        mock_output_handler, _ = self.get_mock_outputs_handler_and_fns()
        mock_tf_outputs_handler.return_value = mock_output_handler

        mock_variable_handler = Mock()
        mock_tf_variables_handler.return_value = mock_variable_handler

        mock_cmd_runner, mock_cmd_runner_fns = self.get_mock_manifest_cmd_runner_and_fns()
        mock_cmd_runner_class.return_value = mock_cmd_runner

        mock_console = mock_console_class.return_value

        # Execute
        handler = OrganizationHandler()
        handler.set_organization("org-name")

        # Verify
        mock_manifest_fns["get_command"].assert_called_once_with("organization.set")
        mock_cmd_runner_class.assert_called_once_with(
            console=mock_console,
            output_handler=mock_output_handler,
            variable_handler=mock_variable_handler,
        )
        mock_cmd_runner_fns["run_command_sequence"].assert_called_once_with(
            mock_cmd,
            cli_paramdefs={
                "organization": StrResolvedCliParameter(parameter_name="organization", value="org-name"),
                "category": StrResolvedCliParameter(parameter_name="category", value="org"),
            },
        )
        mock_cmd_runner_fns["update_variables"].assert_called_once_with(mock_cmd)

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    @patch("jupyter_deploy.engine.terraform.tf_outputs.TerraformOutputsHandler")
    @patch("jupyter_deploy.provider.manifest_command_runner.ManifestCommandRunner")
    @patch("rich.console.Console")
    def test_unset_organization_calls_run_command_sequence_with_correct_params(
        self,
        mock_console_class: Mock,
        mock_cmd_runner_class: Mock,
        mock_tf_outputs_handler: Mock,
        mock_tf_variables_handler: Mock,
        mock_retrieve_manifest: Mock,
    ) -> None:
        # Setup
        mock_cmd = Mock()
        mock_manifest, mock_manifest_fns = self.get_mock_manifest_and_fns()
        mock_manifest_fns["get_command"].return_value = mock_cmd
        mock_retrieve_manifest.return_value = mock_manifest

        mock_output_handler, _ = self.get_mock_outputs_handler_and_fns()
        mock_tf_outputs_handler.return_value = mock_output_handler

        mock_variable_handler = Mock()
        mock_tf_variables_handler.return_value = mock_variable_handler

        mock_cmd_runner, mock_cmd_runner_fns = self.get_mock_manifest_cmd_runner_and_fns()
        mock_cmd_runner_class.return_value = mock_cmd_runner

        mock_console = mock_console_class.return_value

        # Execute
        handler = OrganizationHandler()
        handler.unset_organization()

        # Verify
        mock_manifest_fns["get_command"].assert_called_once_with("organization.unset")
        mock_cmd_runner_class.assert_called_once_with(
            console=mock_console,
            output_handler=mock_output_handler,
            variable_handler=mock_variable_handler,
        )
        mock_cmd_runner_fns["run_command_sequence"].assert_called_once_with(
            mock_cmd,
            cli_paramdefs={
                "category": StrResolvedCliParameter(parameter_name="category", value="org"),
            },
        )
        mock_cmd_runner_fns["update_variables"].assert_called_once_with(mock_cmd)

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    @patch("jupyter_deploy.engine.terraform.tf_outputs.TerraformOutputsHandler")
    @patch("jupyter_deploy.provider.manifest_command_runner.ManifestCommandRunner")
    @patch("rich.console.Console")
    def test_get_organization_calls_run_command_sequence_with_correct_params(
        self,
        mock_console_class: Mock,
        mock_cmd_runner_class: Mock,
        mock_tf_outputs_handler: Mock,
        mock_tf_variables_handler: Mock,
        mock_retrieve_manifest: Mock,
    ) -> None:
        # Setup
        mock_cmd = Mock()
        mock_manifest, mock_manifest_fns = self.get_mock_manifest_and_fns()
        mock_manifest_fns["get_command"].return_value = mock_cmd
        mock_retrieve_manifest.return_value = mock_manifest

        mock_output_handler, _ = self.get_mock_outputs_handler_and_fns()
        mock_tf_outputs_handler.return_value = mock_output_handler

        mock_variable_handler = Mock()
        mock_tf_variables_handler.return_value = mock_variable_handler

        mock_cmd_runner, mock_cmd_runner_fns = self.get_mock_manifest_cmd_runner_and_fns()
        mock_cmd_runner_class.return_value = mock_cmd_runner
        mock_cmd_runner_fns["get_result_value"].return_value = "org-name"

        mock_console = mock_console_class.return_value

        # Execute
        handler = OrganizationHandler()
        result = handler.get_organization()

        # Verify
        self.assertEqual(result, "org-name")
        mock_manifest_fns["get_command"].assert_called_once_with("organization.get")
        mock_cmd_runner_class.assert_called_once_with(
            console=mock_console,
            output_handler=mock_output_handler,
            variable_handler=mock_variable_handler,
        )
        mock_cmd_runner_fns["run_command_sequence"].assert_called_once_with(
            mock_cmd,
            cli_paramdefs={
                "category": StrResolvedCliParameter(parameter_name="category", value="org"),
            },
        )
        mock_cmd_runner_fns["get_result_value"].assert_called_once_with(mock_cmd, "organization.get", str)
        mock_cmd_runner_fns["update_variables"].assert_not_called()

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    @patch("jupyter_deploy.engine.terraform.tf_outputs.TerraformOutputsHandler")
    @patch("jupyter_deploy.provider.manifest_command_runner.ManifestCommandRunner")
    @patch("rich.console.Console")
    def test_organization_methods_do_not_update_variables_when_command_fails(
        self,
        mock_console_class: Mock,
        mock_cmd_runner_class: Mock,
        mock_tf_outputs_handler: Mock,
        mock_tf_variables_handler: Mock,
        mock_retrieve_manifest: Mock,
    ) -> None:
        # Setup
        mock_cmd = Mock()
        mock_manifest, mock_manifest_fns = self.get_mock_manifest_and_fns()
        mock_manifest_fns["get_command"].return_value = mock_cmd
        mock_retrieve_manifest.return_value = mock_manifest

        mock_output_handler, _ = self.get_mock_outputs_handler_and_fns()
        mock_tf_outputs_handler.return_value = mock_output_handler

        mock_variable_handler = Mock()
        mock_tf_variables_handler.return_value = mock_variable_handler

        mock_cmd_runner, mock_cmd_runner_fns = self.get_mock_manifest_cmd_runner_and_fns()
        # Set up run_command_sequence to return False (command failed)
        mock_cmd_runner_fns["run_command_sequence"].return_value = (False, {})
        mock_cmd_runner_class.return_value = mock_cmd_runner

        handler = OrganizationHandler()

        # Test set_organization
        handler.set_organization("org-name")
        mock_cmd_runner_fns["update_variables"].assert_not_called()

        # Reset mocks
        mock_manifest_fns["get_command"].reset_mock()
        mock_cmd_runner_fns["update_variables"].reset_mock()

        # Test unset_organization
        handler.unset_organization()
        mock_cmd_runner_fns["update_variables"].assert_not_called()



from typing import Any, TypeVar, get_origin

# we should remove this import
# the instruction runner should NOT depend on the CLI because
# the CLI should just be one way to run these instructions.
import typer
from rich import console as rich_console

from jupyter_deploy import transform_utils
from jupyter_deploy.engine.engine_outputs import EngineOutputsHandler
from jupyter_deploy.engine.engine_variables import EngineVariablesHandler
from jupyter_deploy.enum import InstructionArgumentSource, ResultSource, UpdateSource
from jupyter_deploy.manifest import JupyterDeployCommandV1
from jupyter_deploy.provider.instruction_runner import InterruptInstructionError
from jupyter_deploy.provider.instruction_runner_factory import InstructionRunnerFactory
from jupyter_deploy.provider.resolved_argdefs import (
    ResolvedInstructionArgument,
    resolve_cliparam_argdef,
    resolve_output_argdef,
    resolve_result_argdef,
)
from jupyter_deploy.provider.resolved_clidefs import ResolvedCliParameter
from jupyter_deploy.provider.resolved_resultdefs import ResolvedInstructionResult

R = TypeVar("R")


class ManifestCommandRunner:
    """Convenience class to run command sequences defined in the project manifest."""

    def __init__(
        self,
        console: rich_console.Console,
        output_handler: EngineOutputsHandler,
        variable_handler: EngineVariablesHandler,
    ) -> None:
        """Instantiate the command runner."""
        self._console = console
        self._output_handler = output_handler
        self._variable_handler = variable_handler
        self._resolved_resultdefs: dict[str, ResolvedInstructionResult] = {}

    def run_command_sequence(
        self, cmd_def: JupyterDeployCommandV1, cli_paramdefs: dict[str, ResolvedCliParameter]
    ) -> tuple[bool, dict[str, ResolvedInstructionResult]]:
        """Execute the cmd, return a tuple of success flag, resolved results."""

        # run all the instructions and collect results
        resolved_resultdefs: dict[str, ResolvedInstructionResult] = {}

        # run instructions
        for instruction_idx, instruction in enumerate(cmd_def.sequence):
            api_name = instruction.api_name
            runner = InstructionRunnerFactory.get_provider_instruction_runner(api_name, self._output_handler)
            output_defs = self._output_handler.get_full_project_outputs()  # cached - okay to call in loop
            resolved_argdefs: dict[str, ResolvedInstructionArgument] = {}

            for arg_def in instruction.arguments:
                arg_name = arg_def.api_attribute
                arg_source_type = arg_def.get_source_type()
                source_key = arg_def.source_key

                if arg_source_type == InstructionArgumentSource.TEMPLATE_OUTPUT:
                    resolved_argdefs[arg_name] = resolve_output_argdef(
                        outdefs=output_defs, arg_name=arg_name, source_key=source_key
                    )
                elif arg_source_type == InstructionArgumentSource.INSTRUCTION_RESULT:
                    resolved_argdefs[arg_name] = resolve_result_argdef(
                        resultdefs=resolved_resultdefs, arg_name=arg_name, source_key=source_key
                    )
                elif arg_source_type == InstructionArgumentSource.CLI_ARGUMENT:
                    resolved_argdefs[arg_name] = resolve_cliparam_argdef(
                        paramdefs=cli_paramdefs, arg_name=arg_name, source_key=source_key
                    )
                else:
                    raise NotImplementedError(f"Argument source is not handled: {arg_source_type}")

            try:
                instruction_results = runner.execute_instruction(
                    instruction_name=api_name,
                    resolved_arguments=resolved_argdefs,
                    console=self._console,
                )
                for instruction_result_name, instruction_result_def in instruction_results.items():
                    indexed_result_name = f"[{instruction_idx}].{instruction_result_name}"
                    resolved_resultdefs[indexed_result_name] = instruction_result_def
            except InterruptInstructionError:
                typer.Abort()
                self._resolved_resultdefs = resolved_resultdefs
                return False, resolved_resultdefs

        self._resolved_resultdefs = resolved_resultdefs
        return True, resolved_resultdefs

    def get_result_value(self, cmd_def: JupyterDeployCommandV1, result_name: str, expect_type: type[R]) -> R:
        """Return the transformed result value."""
        result = next(r for r in cmd_def.results or [] if r.result_name == result_name)

        if not result:
            raise KeyError(f"result '{result_name}' not found in command: {cmd_def.cmd}")

        source_type = result.get_source_type()
        source_key = result.source_key
        transform_type = result.get_transform_type()
        transform_fn = transform_utils.get_transform_fn(transform_type)

        if source_type != ResultSource.INSTRUCTION_RESULT:
            raise NotImplementedError("Invalid type: update only support results from instructions.")
        if source_key not in self._resolved_resultdefs:
            raise KeyError(f"Source-key '{source_key}' not found in result defs for variable: {result_name}.")
        result_def = self._resolved_resultdefs[source_key]

        value = transform_fn(result_def.value)

        base_type = get_origin(expect_type) or expect_type
        if not isinstance(value, base_type):
            raise TypeError(
                f"Expected result '{result_name}' to be of type {expect_type.__name__}, got {type(value).__name__}"
            )
        return value  # type: ignore

    def update_variables(self, cmd_def: JupyterDeployCommandV1) -> None:
        """Update the project variables based on the results."""
        updates = cmd_def.updates

        if not updates:
            return

        varvalues: dict[str, Any] = {}
        for update in updates:
            variable_name = update.variable_name
            source_type = update.get_source_type()
            source_key = update.source_key

            if source_type != UpdateSource.INSTRUCTION_RESULT:
                raise NotImplementedError("Invalid type: update only support results from instructions")
            if source_key not in self._resolved_resultdefs:
                raise KeyError(f"Source-key '{source_key}' not found in result defs for variable: {variable_name}")
            result_def = self._resolved_resultdefs[source_key]

            transform_type = update.get_transform_type()
            transform_fn = transform_utils.get_transform_fn(transform_type)

            value = transform_fn(result_def.value)
            varvalues[variable_name] = value

        self._variable_handler.update_variable_records(varvalues)
        self._variable_handler.sync_project_variables_config(varvalues)



import unittest
from pathlib import Path
from unittest.mock import Mock, patch

from jupyter_deploy.engine.enum import EngineType
from jupyter_deploy.handlers.project.variables_handler import VariablesHandler


class TestVariablesHandler(unittest.TestCase):
    def get_mock_manifest_and_fns(self) -> tuple[Mock, dict[str, Mock]]:
        """Return mocked manifest."""
        mock_manifest = Mock()
        mock_get_engine = Mock()
        mock_manifest.get_engine = mock_get_engine
        mock_get_engine.return_value = EngineType.TERRAFORM

        return (mock_manifest, {"get_engine": mock_get_engine})

    def get_mock_handler_and_fns(self) -> tuple[Mock, dict[str, Mock]]:
        """Return mocked config handler."""
        mock_handler = Mock()
        mock_is_template_directory = Mock()
        mock_get_template_variables = Mock()

        mock_handler.is_template_directory = mock_is_template_directory
        mock_handler.get_template_variables = mock_get_template_variables

        mock_is_template_directory.return_value = True
        mock_get_template_variables.return_value = {}

        return (
            mock_handler,
            {
                "is_template_directory": mock_is_template_directory,
                "get_template_variables": mock_get_template_variables,
            },
        )

    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    def test_handler_implements_all_engines(self, mock_tf_handler: Mock) -> None:
        # Setup
        mock_manifest, _ = self.get_mock_manifest_and_fns()
        tf_mock_handler_instance, _ = self.get_mock_handler_and_fns()
        mock_tf_handler.return_value = tf_mock_handler_instance

        # In the future, if more engines are added, this test should be updated
        # Right now, only Terraform is supported
        handler = VariablesHandler(project_path=Path("/some/cur/dir"), project_manifest=mock_manifest)
        self.assertIsNotNone(handler)

    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    def test_handler_implements_to_tf_engine(self, mock_tf_handler: Mock) -> None:
        # Setup
        path = Path("/some/cur/dir")
        mock_manifest, mock_manifest_fns = self.get_mock_manifest_and_fns()
        tf_mock_handler_instance, tf_fns = self.get_mock_handler_and_fns()
        mock_tf_handler.return_value = tf_mock_handler_instance

        # Execute
        VariablesHandler(path, project_manifest=mock_manifest)

        # Assert
        mock_manifest_fns["get_engine"].assert_called_once()
        mock_tf_handler.assert_called_once_with(project_path=path, project_manifest=mock_manifest)
        tf_fns["is_template_directory"].assert_not_called()
        tf_fns["get_template_variables"].assert_not_called()

    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    def test_handler_calls_underlying_is_template_dir_method(self, mock_tf_handler: Mock) -> None:
        mock_manifest, _ = self.get_mock_manifest_and_fns()
        tf_mock_handler_instance, tf_fns = self.get_mock_handler_and_fns()
        mock_tf_handler.return_value = tf_mock_handler_instance

        handler = VariablesHandler(Path("/some/cur/dir"), mock_manifest)
        result = handler.is_template_directory()

        self.assertTrue(result)
        tf_fns["is_template_directory"].assert_called_once()
        tf_fns["get_template_variables"].assert_not_called()

    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    def test_handler_calls_underlying_get_template_variables_method(self, mock_tf_handler: Mock) -> None:
        mock_manifest, _ = self.get_mock_manifest_and_fns()
        tf_mock_handler_instance, tf_fns = self.get_mock_handler_and_fns()
        mock_tf_handler.return_value = tf_mock_handler_instance

        mock_vars = {"var1": Mock(), "var2": Mock()}
        tf_fns["get_template_variables"].return_value = mock_vars

        handler = VariablesHandler(Path("/some/cur/dir"), mock_manifest)
        result = handler.get_template_variables()

        self.assertEqual(result, mock_vars)
        tf_fns["is_template_directory"].assert_called_once()
        tf_fns["get_template_variables"].assert_called_once()

    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    def test_handler_raises_when_underlying_get_method_raises(self, mock_tf_handler: Mock) -> None:
        mock_manifest, _ = self.get_mock_manifest_and_fns()
        tf_mock_handler_instance, tf_fns = self.get_mock_handler_and_fns()
        mock_tf_handler.return_value = tf_mock_handler_instance
        tf_fns["get_template_variables"].side_effect = RuntimeError()

        handler = VariablesHandler(Path("/some/cur/dir"), mock_manifest)
        with self.assertRaises(RuntimeError):
            handler.get_template_variables()

    @patch("jupyter_deploy.engine.terraform.tf_variables.TerraformVariablesHandler")
    def test_handler_skips_get_method_if_is_template_dir_returns_false(self, mock_tf_handler: Mock) -> None:
        mock_manifest, _ = self.get_mock_manifest_and_fns()
        tf_mock_handler_instance, tf_fns = self.get_mock_handler_and_fns()
        mock_tf_handler.return_value = tf_mock_handler_instance

        tf_fns["is_template_directory"].return_value = False
        mock_vars = {"var1": Mock(), "var2": Mock()}
        tf_fns["get_template_variables"].return_value = mock_vars

        handler = VariablesHandler(Path("/some/cur/dir"), mock_manifest)
        result = handler.get_template_variables()

        self.assertEqual(result, {})
        tf_fns["get_template_variables"].assert_not_called()



import subprocess
import sys
from pathlib import Path
from typing import Annotated

import typer
from jupyter_core.application import JupyterApp
from rich.console import Console

from jupyter_deploy import cmd_utils
from jupyter_deploy.cli.host_app import host_app
from jupyter_deploy.cli.organization_app import organization_app
from jupyter_deploy.cli.servers_app import servers_app
from jupyter_deploy.cli.teams_app import teams_app
from jupyter_deploy.cli.users_app import users_app
from jupyter_deploy.cli.variables_decorator import with_project_variables
from jupyter_deploy.engine.enum import EngineType
from jupyter_deploy.engine.vardefs import TemplateVariableDefinition
from jupyter_deploy.handlers.init_handler import InitHandler
from jupyter_deploy.handlers.project import config_handler
from jupyter_deploy.handlers.project.down_handler import DownHandler
from jupyter_deploy.handlers.project.open_handler import OpenHandler
from jupyter_deploy.handlers.project.show_handler import ShowHandler
from jupyter_deploy.handlers.project.up_handler import UpHandler
from jupyter_deploy.infrastructure.enum import AWSInfrastructureType, InfrastructureType
from jupyter_deploy.provider.enum import ProviderType


class JupyterDeployCliRunner:
    """Wrapper class for Typer app."""

    def __init__(self) -> None:
        """Setup the CLI shell, add sub-commands."""
        self.app = typer.Typer(
            help=("Jupyter-deploy CLI helps you deploy notebooks application to your favorite Cloud provider."),
            no_args_is_help=True,
        )
        self._setup_basic_commands()
        self.app.add_typer(servers_app, name="server")
        self.app.add_typer(users_app, name="users")
        self.app.add_typer(teams_app, name="teams")
        self.app.add_typer(organization_app, name="organization")
        self.app.add_typer(host_app, name="host")

    def _setup_basic_commands(self) -> None:
        """Register the basic commands."""
        pass

    def run(self) -> None:
        """Execute the CLI."""
        self.app()


runner = JupyterDeployCliRunner()


@runner.app.command()
def init(
    path: Annotated[
        str | None,
        typer.Argument(
            help="Path to the directory where jupyter-deploy will create your project files. "
            "Pass '.' to use your current working directory."
        ),
    ] = None,
    engine: Annotated[
        EngineType, typer.Option("--engine", "-E", help="Infrastructure as code software to manage your resources.")
    ] = EngineType.TERRAFORM,
    provider: Annotated[
        ProviderType, typer.Option("--provider", "-P", help="Cloud provider where your resources will be provisioned.")
    ] = ProviderType.AWS,
    infrastructure: Annotated[
        InfrastructureType,
        typer.Option(
            "--infrastructure",
            "-I",
            help="Infrastructure service that your cloud provider will use to provision your resources.",
        ),
    ] = AWSInfrastructureType.EC2,
    template: Annotated[
        str, typer.Option("--template", "-T", help="Base name of the infrastrucuture as code template (e.g., base)")
    ] = "base",
    overwrite: Annotated[
        bool,
        typer.Option(
            "--overwrite",
            "-o",
            help="Overwrite the project directory instead of failing when the directory is not empty.",
        ),
    ] = False,
) -> None:
    """Initialize a project directory containing the specified infrastructure-as-code template.

    Template will be selected based on the provided parameters - the matching
    template package must have already been installed.

    You must specify a project path which must be a directory. If such a directory is not empty,
    the command will fail unless you passed the `--overwrite` or `-o` flag. `--overwrite` will prompt
    for confirmation before deleting existing content.
    """
    if path is None:
        init_help_cmds = ["jupyter", "deploy", "init", "--help"]
        subprocess.run(init_help_cmds)
        return
    project = InitHandler(
        project_dir=path,
        engine=engine,
        provider=provider,
        infrastructure=infrastructure,
        template=template,
    )
    console = Console()

    if not project.may_export_to_project_path():
        if not overwrite:
            console.line()
            console.print(f":x: The directory {project.project_path} is not empty, aborting.", style="red")
            console.line()
            console.print(
                "If you want to overwrite the content of this directory, use the --overwrite option.\n", style="yellow"
            )
            return
        else:
            console.line()
            console.print(f":warning: The target directory {project.abs_project_path} is not empty.", style="yellow")
            console.line()
            console.print(
                "Initiating the project may overwrite your existing files, are you sure you want to proceed?",
                style="yellow",
            )

            overwrite_existing = typer.confirm("")

            if not overwrite_existing:
                console.line()
                console.print(f"Left files under {project.project_path} untouched.\n", style="yellow")
                typer.Abort()
                return

    project.setup()

    console.print(f"Created start-up project files at: {project.project_path.absolute()}", style="bold green")
    console.line()

    if Path.cwd().absolute() != project.project_path.absolute():
        console.print(
            f"Change your working directory to [bold]{project.project_path}[/] "
            "then you can run `[bold cyan]jd config[/]` to configure the project.",
            style="green",
        )
    else:
        console.print("You can now run `[bold cyan]jd config[/]` to configure the project.")
    console.line()
    console.print(
        "This command will use all the defaults specified in the template, unless you override specific variables."
    )
    console.print("The names of the variables depend on the template, use [bold cyan]--help[/] to find them.")
    console.print("To manually set the value of [italic]all[/] variables, use [bold cyan]--defaults none[/].")
    console.line()


@runner.app.command()
@with_project_variables()
def config(
    defaults_preset_name: Annotated[
        str,
        typer.Option(
            "--defaults",
            "-d",
            help="Name of the preset defaults to use: 'all', 'none' or template-specific preset names.",
        ),
    ] = "all",
    record_secrets: Annotated[
        bool,
        typer.Option(
            "--record-secrets",
            "-s",
            help="Record the values of variables marked 'sensitive'.",
        ),
    ] = False,
    reset: Annotated[
        bool, typer.Option("--reset", "-r", help="Delete previously recorded variables and secrets.")
    ] = False,
    skip_verify: Annotated[
        bool, typer.Option("--skip-verify", help="Avoid verifying that the project dependencies are configured.")
    ] = False,
    output_filename: Annotated[
        str | None, typer.Option("--output-filename", "-f", help="Name of the file to store the configuration to.")
    ] = None,
    variables: Annotated[
        dict[str, TemplateVariableDefinition] | None,
        typer.Option("--variables", "-v", help="Will be removed by the decorator."),
    ] = None,
) -> None:
    """Verify the system configuration, prompt inputs and prepare for deployment.

    You must run this command from a jupyter-deploy project directory created with `jd init`.

    The `config` command will remember your variable values so that you do not need to
    specify them again next time you run `config`.

    You can reset these recorded values with `--reset` or `-r`. Sensitive variables do not
    get recorded unless you pass `--record-secrets` or `-s`.

    You can specify where to save the planned change with `--output-file` or `-f`.
    """
    preset_name = None if defaults_preset_name == "none" else defaults_preset_name
    handler = config_handler.ConfigHandler(output_filename=output_filename)

    if not handler.validate_and_set_preset(preset_name=preset_name, will_reset_variables=reset):
        return

    run_verify = not skip_verify
    run_configure = False

    console = Console()

    if reset:
        console.rule("[bold]jupyter-deploy:[/] resetting recorded variables and secrets")
        handler.reset_recorded_variables()
        handler.reset_recorded_secrets()

    if run_verify:
        console.rule("[bold]jupyter-deploy:[/] verifying requirements")
        run_configure = handler.verify_requirements()
    else:
        console.print("[bold]jupyter-deploy:[/] skipping verification of requirements")
        run_configure = True

    if run_configure:
        console.rule("[bold]jupyter-deploy:[/] configuring the project")
        configured = handler.configure(variable_overrides=variables)

        if not configured:
            return

        console.rule("[bold]jupyter-deploy:[/] recording input values")
        handler.record(record_vars=True, record_secrets=record_secrets)

        # finally, display a message to the user if config ignored the template defaults
        # in favor of the recorded variables, with instructions on how to change this behavior.
        if not handler.has_used_preset(preset_name):
            console.line()
            console.print(
                "[bold]jupyter-deploy[/] reused the variables values that you elected previously "
                f"instead of the template preset: [bold cyan]{preset_name}[/]."
            )
            console.print("You can use `[bold cyan]--reset[/]` to clear your recorded values.")
        console.rule()

        console.print("Your project is ready.", style="bold green")
        console.line()
        if output_filename:
            console.print(
                f"You can now run `[bold cyan]jd up --config-filename {output_filename}[/]` to create the resources."
            )
        else:
            console.print("You can now run `[bold cyan]jd up[/]` to create or update the resources.")
        console.line()


@runner.app.command()
def up(
    project_dir: Annotated[
        str | None, typer.Option("--path", "-p", help="Directory of the jupyter-deploy project to bring up.")
    ] = None,
    config_filename: Annotated[
        str | None,
        typer.Option(
            "--config-filename", "-f", help="Name of a file in the project_dir containing the execution configuration."
        ),
    ] = None,
    auto_approve: Annotated[
        bool, typer.Option("--answer-yes", "-y", help="Apply changes without confirmation prompt.")
    ] = False,
) -> None:
    """Apply the changes defined in the infrastructure-as-code template.

    Run either from a jupyter-deploy project directory that you created with `jd init`;
    or pass a --path PATH to such a directory. Optionally, you can also pass a --config-file
    argument.

    Call `jd config` first to set the input variables and
    verify the configuration.
    """
    with cmd_utils.project_dir(project_dir):
        handler = UpHandler()
        console = Console()

        console.rule("[bold]jupyter-deploy:[/] verifying presence of config file")
        config_file_path = handler.get_config_file_path(config_filename)
        if config_file_path:
            console.rule("[bold]jupyter-deploy:[/] applying infrastructure changes")
            handler.apply(config_file_path, auto_approve)


@runner.app.command()
def down(
    project_dir: Annotated[
        str | None, typer.Option("--path", "-p", help="Directory of the jupyter-deploy project to bring down.")
    ] = None,
    auto_approve: Annotated[
        bool, typer.Option("--answer-yes", "-y", help="Destroy resources without confirmation prompt.")
    ] = False,
) -> None:
    """Destroy the resources defined in the infrastructure-as-code template.

    Run either from a jupyter-deploy project directed that you created with `jd init`;
    or pass a --path PATH to such a directory.

    No-op if you have not already created the infrastructure with `jd up`, or if you
    already ran `jd down`.
    """
    with cmd_utils.project_dir(project_dir):
        handler = DownHandler()
        console = Console()

        console.rule("[bold]jupyter-deploy:[/] destroying infrastructure resources")
        handler.destroy(auto_approve)


@runner.app.command()
def open(
    project_dir: Annotated[
        str | None, typer.Option("--path", "-p", help="Directory of the jupyter-deploy project to open.")
    ] = None,
) -> None:
    """Open the Jupyter app in your webbrowser.

    Run either from a jupyter-deploy project directory that you created with `jd init`;
    or pass a --path PATH to such a directory.

    Call `jd config` and `jd up` first.
    """
    with cmd_utils.project_dir(project_dir):
        handler = OpenHandler()
        handler.open()


@runner.app.command()
def show(
    project_dir: Annotated[
        str | None, typer.Option("--path", "-p", help="Directory of the jupyter-deploy project to show information.")
    ] = None,
    info: Annotated[bool, typer.Option("--info", help="Display core project and template information.")] = False,
    outputs: Annotated[bool, typer.Option("--outputs", help="Display outputs information.")] = False,
    variables: Annotated[bool, typer.Option("--variables", help="Display variables information.")] = False,
) -> None:
    """Display information about the jupyter-deploy project.

    Run either from a jupyter-deploy project directory that you created with `jd init`;
    or pass a --path PATH to such a directory.

    If the project is up, shows the values of the output as defined in
    the infrastructure as code project.
    """
    if not info and not outputs and not variables:
        show_info = True
        show_outputs = True
        show_variables = True
    else:
        show_info = info
        show_outputs = outputs
        show_variables = variables

    with cmd_utils.project_dir(project_dir):
        handler = ShowHandler()
        handler.show_project_info(show_info=show_info, show_outputs=show_outputs, show_variables=show_variables)


class JupyterDeployApp(JupyterApp):
    """Jupyter Deploy application for use with 'jupyter deploy' command."""

    name = "jupyter-deploy"
    description = "Deploy Jupyter notebooks application to your favorite Cloud provider."

    def start(self) -> None:
        """Run the deploy application."""
        args_without_command = sys.argv[2:] if len(sys.argv) > 2 else []
        sys.argv = args_without_command

        runner.run()


def main() -> None:
    if sys.argv[0].endswith("jupyter") and len(sys.argv) > 1 and sys.argv[1] == "deploy":
        JupyterDeployApp.launch_instance()
    else:
        runner.run()



from enum import Enum

import boto3
from mypy_boto3_ec2.client import EC2Client
from rich import console as rich_console

from jupyter_deploy.api.aws.ec2 import ec2_instance
from jupyter_deploy.provider.instruction_runner import InstructionRunner, InterruptInstructionError
from jupyter_deploy.provider.resolved_argdefs import (
    ResolvedInstructionArgument,
    StrResolvedInstructionArgument,
    require_arg,
)
from jupyter_deploy.provider.resolved_resultdefs import ResolvedInstructionResult, StrResolvedInstructionResult


class AwsEc2Instruction(str, Enum):
    """AWS EC2 instructions accessible from manifest.commands[].sequence[].api-name."""

    DESCRIBE_INSTANCE_STATUS = "describe-instance-status"
    START_INSTANCE = "start-instance"
    STOP_INSTANCE = "stop-instance"
    REBOOT_INSTANCE = "reboot-instance"
    WAIT_FOR_RUNNING = "wait-for-running"
    WAIT_FOR_STOPPED = "wait-for-stopped"


class AwsEc2Runner(InstructionRunner):
    """Runner class for AWS EC2 service API instructions."""

    client: EC2Client

    def __init__(self, region_name: str | None) -> None:
        """Instantiates the EC2 boto3 client."""
        self.client: EC2Client = boto3.client("ec2", region_name=region_name)

    def _describe_instance_status(
        self,
        resolved_arguments: dict[str, ResolvedInstructionArgument],
        console: rich_console.Console,
    ) -> dict[str, ResolvedInstructionResult]:
        # retrieve required parameters
        instance_id_arg = require_arg(resolved_arguments, "instance_id", StrResolvedInstructionArgument)
        instance_id = instance_id_arg.value

        console.print(f"Retrieving status of instance: [bold cyan]{instance_id}[/]")
        instance_status = ec2_instance.describe_instance_status(
            self.client,
            instance_id=instance_id,
        )
        console.print(instance_status)
        return {
            "InstanceStateName": StrResolvedInstructionResult(
                result_name="InstanceStateName",
                value=instance_status.get("InstanceState", {}).get("Name", "unknown"),
            )
        }

    def _start_instance(
        self,
        resolved_arguments: dict[str, ResolvedInstructionArgument],
        console: rich_console.Console,
    ) -> dict[str, ResolvedInstructionResult]:
        # retrieve required parameters
        instance_id_arg = require_arg(resolved_arguments, "instance_id", StrResolvedInstructionArgument)
        instance_id = instance_id_arg.value

        instance_status = ec2_instance.describe_instance_status(self.client, instance_id=instance_id)
        state = ec2_instance.Ec2InstanceState.from_state_response(instance_status.get("InstanceState", {}))

        if state == ec2_instance.Ec2InstanceState.PENDING:
            console.print(f":warning: Instance [bold]{instance_id}[/] is already starting...", style="yellow")
            console.line()
            console.print("Wait for the instance to come online.")
            raise InterruptInstructionError
        elif state == ec2_instance.Ec2InstanceState.RUNNING:
            console.print(f":white_check_mark: Instance [bold]{instance_id}[/] is already running.", style="green")
            console.line()
            raise InterruptInstructionError
        elif state == ec2_instance.Ec2InstanceState.SHUTTING_DOWN:
            console.print(f":x: Cannot start instance [bold]{instance_id}[/], it is being terminated.", style="red")
            console.line()
            raise InterruptInstructionError
        elif state == ec2_instance.Ec2InstanceState.TERMINATED:
            console.print(f":x: Cannot start terminated instance [bold]{instance_id}[/].", style="red")
            console.line()
            raise InterruptInstructionError
        elif state == ec2_instance.Ec2InstanceState.STOPPING:
            console.print(f":x: Cannot start stopping instance [bold]{instance_id}[/]...", style="red")
            console.line()
            console.print("Wait for instance to fully stop.")
            raise InterruptInstructionError
        elif not state.is_startable():
            raise ValueError(f"Unhandled instance state: '{state.value}'")

        ec2_instance.start_instance(
            self.client,
            instance_id=instance_id_arg.value,
        )
        console.print(f"Starting instance [bold]{instance_id}[/]...", style="green")
        console.line()
        return {}

    def _stop_instance(
        self,
        resolved_arguments: dict[str, ResolvedInstructionArgument],
        console: rich_console.Console,
    ) -> dict[str, ResolvedInstructionResult]:
        # retrieve required parameters
        instance_id_arg = require_arg(resolved_arguments, "instance_id", StrResolvedInstructionArgument)
        instance_id = instance_id_arg.value

        instance_status = ec2_instance.describe_instance_status(self.client, instance_id=instance_id)
        state = ec2_instance.Ec2InstanceState.from_state_response(instance_status.get("InstanceState", {}))

        if state == ec2_instance.Ec2InstanceState.PENDING:
            console.print(f":x: Instance [bold]{instance_id}[/] is starting...", style="yellow")
            console.line()
            console.print("Wait for the instance to come online.")
            raise InterruptInstructionError
        elif state == ec2_instance.Ec2InstanceState.SHUTTING_DOWN:
            console.print(f":x: Cannot stop instance [bold]{instance_id}[/], it is being terminated.", style="red")
            console.line()
            raise InterruptInstructionError
        elif state == ec2_instance.Ec2InstanceState.TERMINATED:
            console.print(f":x: Cannot stop terminated instance [bold]{instance_id}[/].", style="red")
            console.line()
            raise InterruptInstructionError
        elif state == ec2_instance.Ec2InstanceState.STOPPING:
            console.print(f":warning: Instance [bold]{instance_id}[/] is already stopping...", style="yellow")
            console.line()
            console.print("Wait for the instance to fully stop.")
            raise InterruptInstructionError
        elif state == ec2_instance.Ec2InstanceState.STOPPED:
            console.print(f":white_check_mark: Instance [bold]{instance_id}[/] is already stopped.", style="green")
            console.line()
            raise InterruptInstructionError
        elif not state.is_stoppable():
            raise ValueError(f"Unhandled instance state: '{state.value}'")

        ec2_instance.stop_instance(
            self.client,
            instance_id=instance_id,
        )
        console.print(f"Instance [bold]{instance_id}[/] is stopping...", style="green")
        console.line()
        return {}

    def _reboot_instance(
        self,
        resolved_arguments: dict[str, ResolvedInstructionArgument],
        console: rich_console.Console,
    ) -> dict[str, ResolvedInstructionResult]:
        # retrieve required parameters
        instance_id_arg = require_arg(resolved_arguments, "instance_id", StrResolvedInstructionArgument)
        instance_id = instance_id_arg.value

        instance_status = ec2_instance.describe_instance_status(self.client, instance_id=instance_id)
        state = ec2_instance.Ec2InstanceState.from_state_response(instance_status.get("InstanceState", {}))

        if state == ec2_instance.Ec2InstanceState.PENDING:
            console.print(f":x: Cannot reboot instance [bold]{instance_id}[/], it is still starting...", style="red")
            console.line()
            console.print("Wait for the instance to come online.")
            raise InterruptInstructionError
        elif state == ec2_instance.Ec2InstanceState.SHUTTING_DOWN:
            console.print(f":x: Cannot reboot instance [bold]{instance_id}[/], it is being terminated.", style="red")
            console.line()
            raise InterruptInstructionError
        elif state == ec2_instance.Ec2InstanceState.TERMINATED:
            console.print(f":x: Cannot reboot terminated instance [bold]{instance_id}[/].", style="red")
            console.line()
            raise InterruptInstructionError
        elif state == ec2_instance.Ec2InstanceState.STOPPING:
            console.print(f"Cannot reboot stopping instance [bold]{instance_id}[/].", style="red")
            console.line()
            console.print("Wait for the instance to fully stop, then run `jd host start`.")
            raise InterruptInstructionError
        elif state == ec2_instance.Ec2InstanceState.STOPPED:
            console.print(f":x: Cannot reboot stopped instance [bold]{instance_id}[/].", style="red")
            console.line()
            console.print("Run the start command instead.")
            raise InterruptInstructionError
        elif not state.is_stoppable():
            raise ValueError(f"Unhandled instance state: '{state.value}'")

        ec2_instance.restart_instance(
            self.client,
            instance_id=instance_id,
        )
        console.print(f"Instance [bold]{instance_id}[/] is rebooting...", style="green")
        console.line()
        return {}

    def _wait_for_state(
        self,
        resolved_arguments: dict[str, ResolvedInstructionArgument],
        console: rich_console.Console,
        desired_state: ec2_instance.Ec2InstanceState,
        timeout_seconds: int = 60,
    ) -> dict[str, ResolvedInstructionResult]:
        instance_id_arg = require_arg(resolved_arguments, "instance_id", StrResolvedInstructionArgument)
        instance_id = instance_id_arg.value
        instance_status = ec2_instance.poll_for_instance_status(
            self.client,
            console=console,
            instance_id=instance_id,
            desired_state=desired_state,
            timeout_seconds=timeout_seconds,
        )
        return {
            "InstanceStateName": StrResolvedInstructionResult(
                result_name="InstanceStateName",
                value=instance_status.get("InstanceState", {}).get("Name", "unknown"),
            )
        }

    def execute_instruction(
        self,
        instruction_name: str,
        resolved_arguments: dict[str, ResolvedInstructionArgument],
        console: rich_console.Console,
    ) -> dict[str, ResolvedInstructionResult]:
        if instruction_name == AwsEc2Instruction.DESCRIBE_INSTANCE_STATUS:
            return self._describe_instance_status(
                resolved_arguments=resolved_arguments,
                console=console,
            )
        elif instruction_name == AwsEc2Instruction.START_INSTANCE:
            return self._start_instance(
                resolved_arguments=resolved_arguments,
                console=console,
            )
        elif instruction_name == AwsEc2Instruction.STOP_INSTANCE:
            return self._stop_instance(
                resolved_arguments=resolved_arguments,
                console=console,
            )
        elif instruction_name == AwsEc2Instruction.REBOOT_INSTANCE:
            return self._reboot_instance(
                resolved_arguments=resolved_arguments,
                console=console,
            )
        elif instruction_name == AwsEc2Instruction.WAIT_FOR_RUNNING:
            return self._wait_for_state(
                resolved_arguments=resolved_arguments,
                console=console,
                desired_state=ec2_instance.Ec2InstanceState.RUNNING,
                timeout_seconds=60,  # EC2:StartInstances is generally fast
            )
        elif instruction_name == AwsEc2Instruction.WAIT_FOR_STOPPED:
            return self._wait_for_state(
                resolved_arguments=resolved_arguments,
                console=console,
                desired_state=ec2_instance.Ec2InstanceState.STOPPED,
                timeout_seconds=600,  # GPU instances take a while to stop
            )

        raise NotImplementedError(f"No execution implementation for command: 'aws.ec2.{instruction_name}'")



import unittest
from collections.abc import Callable
from unittest.mock import Mock, patch

from packaging.version import Version

from jupyter_deploy import verify_utils
from jupyter_deploy.enum import JupyterDeployTool


class TestVerifyInstallation(unittest.TestCase):
    @patch("jupyter_deploy.cmd_utils.check_executable_installation")
    def test_returns_true_when_check_exec_installation(self, mock_check: Mock) -> None:
        # Mock the check_executable_installation to return successful installation
        mock_check.return_value = (True, "2.0.0", None)
        result = verify_utils._check_installation("my-tool")

        # Verify
        self.assertTrue(result)
        mock_check.assert_called_once_with(executable_name="my-tool")

    @patch("jupyter_deploy.cmd_utils.check_executable_installation")
    def test_returns_false_when_check_exec_installation_not_found(self, mock_check: Mock) -> None:
        # Mock the check_executable_installation to return failed installation
        mock_check.return_value = (False, None, "Command 'my-tool' not found")
        result = verify_utils._check_installation("my-tool")

        # Verify
        self.assertFalse(result)
        mock_check.assert_called_once()

    @patch("jupyter_deploy.cmd_utils.check_executable_installation")
    def test_returns_true_with_version_check(self, mock_check: Mock) -> None:
        # Mock the check_executable_installation to return successful installation with version
        mock_check.return_value = (True, "2.5.0", None)

        # Call the function under test with a minimum version requirement
        result = verify_utils._check_installation("my-tool", min_version=Version("2.0.0"))

        # Verify
        self.assertTrue(result)
        mock_check.assert_called_once()

    @patch("jupyter_deploy.cmd_utils.check_executable_installation")
    def test_returns_false_when_version_check_fails(self, mock_check: Mock) -> None:
        mock_check.return_value = (True, "1.5.0", None)
        result = verify_utils._check_installation("my-tool", min_version=Version("2.0.0"))

        # Verify
        self.assertFalse(result)
        mock_check.assert_called_once()

    @patch("jupyter_deploy.cmd_utils.check_executable_installation")
    def test_raises_when_check_exec_raises(self, mock_check: Mock) -> None:
        # Mock the check_executable_installation to raise an exception
        mock_check.side_effect = Exception("Test exception")
        with self.assertRaises(Exception) as context:
            verify_utils._check_installation("my-tool")

        # Verify
        self.assertEqual(str(context.exception), "Test exception")
        mock_check.assert_called_once()


class TestVerifyMap(unittest.TestCase):
    def test_all_mapped_to_distinct_verification_method(self) -> None:
        methods: set[Callable] = set()

        for tool, verify_fn in verify_utils._TOOL_VERIFICATION_FN_MAP.items():
            with self.subTest(tool=tool, verify_fn=verify_fn):
                if verify_fn in methods:
                    self.fail(f"duplicate tool verification method for tool: {tool}")

        mapped_tools = set(verify_utils._TOOL_VERIFICATION_FN_MAP.keys())
        for tool in JupyterDeployTool:
            with self.subTest(tool=tool):
                self.assertIn(tool, mapped_tools, f"no verification function mapped for tool: {tool}")

    @patch("jupyter_deploy.verify_utils._check_installation")
    def test_all_verification_methods_should_pass_response(self, mock_check: Mock) -> None:
        for tool, verify_fn in verify_utils._TOOL_VERIFICATION_FN_MAP.items():
            mock_check.reset_mock()
            mock_check.return_value = True
            with self.subTest(tool=tool, verify_fn=verify_fn):
                self.assertTrue(verify_fn(None))
                mock_check.assert_called_once()


class TestVerifyToolsInstallation(unittest.TestCase):
    def test_return_true_for_empty_list(self) -> None:
        self.assertTrue(verify_utils.verify_tools_installation([]))

    @patch("jupyter_deploy.verify_utils._check_installation")
    def test_call_each_verification_method(self, mock_check: Mock) -> None:
        req1 = Mock()
        req2 = Mock()
        req1.name = "aws-cli"
        req1.version = None
        req2.name = "terraform"
        req2.version = None

        mock_check.return_value = True

        self.assertTrue(verify_utils.verify_tools_installation([req1, req2]))
        self.assertEqual(mock_check.call_count, 2)
        self.assertEqual(mock_check.mock_calls[0][2]["min_version"], None)

    @patch("jupyter_deploy.verify_utils._check_installation")
    def test_passes_min_versions(self, mock_check: Mock) -> None:
        req1 = Mock()
        req2 = Mock()
        req1.name = "aws-cli"
        req1.version = "1.0.0"
        req2.name = "terraform"
        req2.version = "2.0.0"

        mock_check.side_effect = [False, True]

        self.assertFalse(verify_utils.verify_tools_installation([req1, req2]))
        self.assertEqual(mock_check.call_count, 2)
        self.assertEqual(mock_check.mock_calls[0][2]["min_version"], Version("1.0.0"))
        self.assertEqual(mock_check.mock_calls[1][2]["min_version"], Version("2.0.0"))

    @patch("jupyter_deploy.verify_utils._check_installation")
    def test_skips_unrecognized_tools(self, mock_check: Mock) -> None:
        req1 = Mock()
        req2 = Mock()
        req3 = Mock()
        req1.name = "aws-cli"
        req1.version = None
        req2.name = "some-unknown-tool"
        req2.version = None
        req3.name = "jq"
        req3.version = None

        mock_check.return_value = True

        self.assertTrue(verify_utils.verify_tools_installation([req1, req2, req3]))
        self.assertEqual(mock_check.call_count, 2)

    @patch("jupyter_deploy.verify_utils._check_installation")
    def test_skip_version_verification_for_invalid_versions(self, mock_check: Mock) -> None:
        req1 = Mock()
        req2 = Mock()
        req1.name = "aws-cli"
        req1.version = "i-am-not-a-version"
        req2.name = "terraform"
        req2.version = None

        mock_check.return_value = False

        self.assertFalse(verify_utils.verify_tools_installation([req1, req2]))
        self.assertEqual(mock_check.call_count, 2)
        self.assertEqual(mock_check.mock_calls[0][2]["min_version"], None)
        self.assertEqual(mock_check.mock_calls[1][2]["min_version"], None)



import unittest
from pathlib import Path
from unittest.mock import Mock, patch

from jupyter_deploy.handlers.project.up_handler import UpHandler
from jupyter_deploy.manifest import JupyterDeployManifestV1


class TestUpHandler(unittest.TestCase):
    def setUp(self) -> None:
        self.mock_manifest = JupyterDeployManifestV1(
            **{  # type: ignore
                "schema_version": 1,
                "template": {
                    "name": "mock-template-name",
                    "engine": "terraform",
                    "version": "1.0.0",
                },
            }
        )

    @patch("jupyter_deploy.engine.terraform.tf_up.TerraformUpHandler")
    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("pathlib.Path.cwd")
    def test_init_creates_terraform_handler(
        self, mock_cwd: Mock, mock_retrieve_manifest: Mock, mock_tf_handler_cls: Mock
    ) -> None:
        mock_cwd.return_value = Path("/mock/cwd")
        mock_retrieve_manifest.return_value = self.mock_manifest
        mock_tf_handler = Mock()
        mock_tf_handler_cls.return_value = mock_tf_handler
        mock_tf_handler.engine_dir_path = Path("/mock/cwd/engine")

        handler = UpHandler()

        mock_tf_handler_cls.assert_called_once_with(project_path=Path("/mock/cwd"))
        self.assertEqual(handler._handler, mock_tf_handler)

    @patch("jupyter_deploy.engine.terraform.tf_up.TerraformUpHandler")
    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("pathlib.Path.cwd")
    def test_apply_delegates_to_handler(
        self, mock_cwd: Mock, mock_retrieve_manifest: Mock, mock_tf_handler_cls: Mock
    ) -> None:
        path = Path("/mock/path")
        mock_cwd.return_value = Path("/mock/cwd")
        mock_retrieve_manifest.return_value = self.mock_manifest
        mock_tf_handler = Mock()
        mock_tf_handler_cls.return_value = mock_tf_handler
        mock_tf_handler.engine_dir_path = Path("/mock/cwd/engine")

        handler = UpHandler()
        handler.apply(path, auto_approve=False)

        mock_tf_handler.apply.assert_called_once_with(path, False)

    @patch("jupyter_deploy.engine.terraform.tf_up.TerraformUpHandler")
    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("pathlib.Path.cwd")
    def test_apply_propagates_exceptions(
        self, mock_cwd: Mock, mock_retrieve_manifest: Mock, mock_tf_handler_cls: Mock
    ) -> None:
        path = Path("/mock/path")
        mock_cwd.return_value = Path("/mock/cwd")
        mock_retrieve_manifest.return_value = self.mock_manifest
        mock_tf_handler = Mock()
        mock_tf_handler.apply.side_effect = Exception("Apply failed")
        mock_tf_handler.engine_dir_path = Path("/mock/cwd/engine")
        mock_tf_handler_cls.return_value = mock_tf_handler

        handler = UpHandler()

        with self.assertRaises(Exception) as context:
            handler.apply(path)

        self.assertEqual(str(context.exception), "Apply failed")
        mock_tf_handler.apply.assert_called_once()

    @patch("jupyter_deploy.engine.terraform.tf_up.TerraformUpHandler")
    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("pathlib.Path.cwd")
    def test_get_default_config_filename_delegates_to_handler(
        self, mock_cwd: Mock, mock_retrieve_manifest: Mock, mock_tf_handler_cls: Mock
    ) -> None:
        mock_cwd.return_value = Path("/mock/cwd")
        mock_retrieve_manifest.return_value = self.mock_manifest
        mock_tf_handler = Mock()
        mock_tf_handler.engine_dir_path = Path("/mock/cwd/engine")
        mock_tf_handler.get_default_config_filename.return_value = "jdout-tfplan"
        mock_tf_handler_cls.return_value = mock_tf_handler

        handler = UpHandler()
        result = handler.get_default_config_filename()

        mock_tf_handler.get_default_config_filename.assert_called_once()
        self.assertEqual(result, "jdout-tfplan")

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("pathlib.Path.cwd")
    def test_init_raises_value_error_for_unsupported_engine(self, mock_cwd: Mock, mock_retrieve_manifest: Mock) -> None:
        mock_cwd.return_value = Path("/mock/cwd")
        mock_manifest = self.mock_manifest
        # Change the engine to an unsupported one
        mock_manifest.template.engine = "UNSUPPORTED_ENGINE"  # type: ignore
        mock_retrieve_manifest.return_value = mock_manifest

        with self.assertRaises(ValueError):
            UpHandler()

    @patch("jupyter_deploy.engine.terraform.tf_up.TerraformUpHandler")
    @patch("jupyter_deploy.handlers.project.up_handler.Console")
    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("pathlib.Path.cwd")
    def test_get_config_file_path_when_file_exists(
        self,
        mock_cwd: Mock,
        mock_retrieve_manifest: Mock,
        mock_console_cls: Mock,
        mock_tf_handler_cls: Mock,
    ) -> None:
        mock_cwd.return_value = Path("/mock/cwd")
        mock_retrieve_manifest.return_value = self.mock_manifest
        mock_tf_handler = Mock()
        mock_tf_handler.engine_dir_path = Path("/mock/cwd/engine")
        mock_tf_handler.get_default_config_filename.return_value = "jdout-tfplan"
        mock_tf_handler_cls.return_value = mock_tf_handler

        with patch.object(Path, "exists", return_value=True):
            handler = UpHandler()
            result = handler.get_config_file_path("test-config")

        self.assertEqual(result, Path("/mock/cwd/engine/test-config"))

    @patch("jupyter_deploy.engine.terraform.tf_up.TerraformUpHandler")
    @patch("pathlib.Path")
    @patch("jupyter_deploy.handlers.project.up_handler.Console")
    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("pathlib.Path.cwd")
    def test_get_config_file_path_when_file_does_not_exist(
        self,
        mock_cwd: Mock,
        mock_retrieve_manifest: Mock,
        mock_console_cls: Mock,
        mock_path_cls: Mock,
        mock_tf_handler_cls: Mock,
    ) -> None:
        mock_cwd.return_value = Path("/mock/cwd")
        mock_retrieve_manifest.return_value = self.mock_manifest
        config_path = Path("/mock/cwd/test-config")

        mock_path_cls.return_value = config_path
        mock_tf_handler = Mock()
        mock_tf_handler.engine_dir_path = Path("/mock/cwd/engine")
        mock_tf_handler.get_default_config_filename.return_value = "jdout-tfplan"
        mock_tf_handler_cls.return_value = mock_tf_handler
        mock_console_instance = Mock()
        mock_console_cls.return_value = mock_console_instance

        with patch.object(Path, "exists", return_value=False):
            handler = UpHandler()
            result = handler.get_config_file_path("test-config")

        self.assertEqual(result, Path("/mock/cwd"))
        mock_console_instance.print.assert_called_once()



import unittest
from pathlib import Path
from typing import cast
from unittest.mock import patch

from jupyter_deploy.engine.terraform import tf_outdefs, tf_outfiles
from jupyter_deploy.engine.terraform.tf_types import TerraformType


class TestExtractDescriptionFromDotTfFile(unittest.TestCase):
    outputs_tf_content: str

    @classmethod
    def setUpClass(cls) -> None:
        mock_outputs_path = Path(__file__).parent / "mock_outputs.tf"
        with open(mock_outputs_path) as f:
            cls.outputs_tf_content = f.read()

    def test_return_empty_dict_on_empty_content(self) -> None:
        result = tf_outfiles.extract_description_from_dot_tf_content("")
        self.assertEqual({}, result)

    def test_parsing_actual_outputs_dot_tf(self) -> None:
        content = self.outputs_tf_content
        result = tf_outfiles.extract_description_from_dot_tf_content(content)

        expected = {"instance_id": "ID of the jupyter notebook.", "aws_region": "Name of the AWS region."}
        self.assertEqual(expected, result)


class TestCombineCmdAndOutputsDotTfResults(unittest.TestCase):
    def test_combine_add_description_to_matched(self) -> None:
        # Setup
        output_def = tf_outdefs.TerraformStrOutputDefinition(
            output_name="instance_id", value="i-12345abcdef", tf_type=TerraformType.STRING
        )
        output_defs_from_cmd: dict[str, tf_outdefs.TerraformOutputDefinition] = {
            "instance_id": cast(tf_outdefs.TerraformOutputDefinition, output_def)
        }
        descriptions_from_file = {"instance_id": "ID of the jupyter notebook."}

        # Execute
        result = tf_outfiles.combine_cmd_and_outputs_dot_tf_results(output_defs_from_cmd, descriptions_from_file)

        # Verify
        self.assertEqual(1, len(result))
        self.assertEqual("ID of the jupyter notebook.", result["instance_id"].description)

    def test_combine_add_empty_description_if_not_found_in_outputs_dot_tf(self) -> None:
        # Setup
        output_def = tf_outdefs.TerraformStrOutputDefinition(
            output_name="missing_output", value="some-value", tf_type=TerraformType.STRING
        )
        output_defs_from_cmd: dict[str, tf_outdefs.TerraformOutputDefinition] = {
            "missing_output": cast(tf_outdefs.TerraformOutputDefinition, output_def)
        }
        descriptions_from_file = {"instance_id": "ID of the jupyter notebook."}

        # Execute
        with patch("builtins.print") as mock_print:
            result = tf_outfiles.combine_cmd_and_outputs_dot_tf_results(output_defs_from_cmd, descriptions_from_file)

        # Verify
        self.assertEqual(1, len(result))
        self.assertEqual("", result["missing_output"].description)
        mock_print.assert_called_once_with("Warning: output 'missing_output' not found in outputs.tf file.")

    def test_combine_ignore_output_dot_tf_entries_not_matching(self) -> None:
        # Setup
        output_def = tf_outdefs.TerraformStrOutputDefinition(
            output_name="instance_id", value="i-12345abcdef", tf_type=TerraformType.STRING
        )
        output_defs_from_cmd: dict[str, tf_outdefs.TerraformOutputDefinition] = {
            "instance_id": cast(tf_outdefs.TerraformOutputDefinition, output_def)
        }
        descriptions_from_file = {
            "instance_id": "ID of the jupyter notebook.",
            "extra_output": "This output is not in cmd results.",
        }

        # Execute
        result = tf_outfiles.combine_cmd_and_outputs_dot_tf_results(output_defs_from_cmd, descriptions_from_file)

        # Verify
        self.assertEqual(1, len(result))
        self.assertEqual("ID of the jupyter notebook.", result["instance_id"].description)
        self.assertNotIn("extra_output", result)

    def test_combine_returns_empty_dict_on_empty_cmd_defs(self) -> None:
        # Setup
        output_defs_from_cmd: dict[str, tf_outdefs.TerraformOutputDefinition] = {}
        descriptions_from_file: dict[str, str] = {}

        # Execute
        result = tf_outfiles.combine_cmd_and_outputs_dot_tf_results(output_defs_from_cmd, descriptions_from_file)

        # Verify
        self.assertEqual({}, result)



import unittest

from parameterized import parameterized  # type: ignore

from jupyter_deploy.str_utils import get_trimmed_header, to_cli_option_name, to_list_str


class TestToCliOptionName(unittest.TestCase):
    @parameterized.expand(
        [
            ("FullTitleCase", "full-title-case"),
            ("camelCaseName", "camel-case-name"),
            ("python_var_name", "python-var-name"),
            ("SomeMixed-Case", "some-mixed-case"),
            ("XmlHttpRequest", "xml-http-request"),
            ("API_KEY", "api-key"),
            ("already-kebab-case", "already-kebab-case"),
            ("UPPERCASE", "uppercase"),
            ("lowercase", "lowercase"),
            ("snake_case_variable", "snake-case-variable"),
            ("Mixed_snake_Case", "mixed-snake-case"),
            ("With-Existing-Hyphens", "with-existing-hyphens"),
            ("Multiple___Underscores", "multiple-underscores"),
            ("Multiple---dashes", "multiple-dashes"),
            ("CamelCaseWithNUMBER123", "camel-case-with-number123"),
        ]
    )
    def test_valid_values(self, input_str: str, expect_str: str) -> None:
        result = to_cli_option_name(input_str)
        self.assertEqual(result, expect_str)

    def test_empty_string_does_not_raise(self) -> None:
        result = to_cli_option_name("")
        self.assertEqual(result, "")


class TestTrimmedHeader(unittest.TestCase):
    def test_actual_tf_variables_description(self) -> None:
        full_desc = (
            "      Client ID of the OAuth app that will control access to your jupyter notebooks.\n"
            "\n"
            "    You must create an OAuth app first in your Github account.\n"
            "    1. Open GitHub: https://github.com/\n"
            "    2. Select your user icon on the top right\n"
        )

        expected = "Client ID of the OAuth app that will control access to your jupyter notebooks."
        result = get_trimmed_header(full_desc)
        self.assertEqual(result, expected)

    def test_trim_leading_spaces(self) -> None:
        input_text = "    This has leading spaces"
        expected = "This has leading spaces"

        result = get_trimmed_header(input_text)
        self.assertEqual(result, expected)

    def test_keeps_only_first_line(self) -> None:
        input_text = "First line\nSecond line\nThird line"
        expected = "First line"

        result = get_trimmed_header(input_text)
        self.assertEqual(result, expected)

    def test_trim_first_line_if_too_long(self) -> None:
        long_text = (
            "This is a very long line that exceeds the default maximum length of 120 characters. "
            "It should be trimmed to exactly 120 characters when processed by the get_trimmed_header function."
        )
        max_length = 30
        result = get_trimmed_header(long_text, max_length)
        self.assertEqual(len(result), max_length)
        self.assertEqual(result, long_text[:max_length])

    def test_empty_string_does_not_raise(self) -> None:
        result = get_trimmed_header("")
        self.assertEqual(result, "")

    @parameterized.expand(
        [
            ("Simple text", 120, "Simple text"),
            ("  Indented text", 120, "Indented text"),
            ("First line\nSecond line", 120, "First line"),
            ("", 120, ""),
            ("Short text", 5, "Short"),
            ("\n\nEmpty lines before text", 120, "Empty lines before text"),
            ("Text with\ttabs", 120, "Text with\ttabs"),
            ("   Mixed   spaces   ", 120, "Mixed   spaces   "),
            ("Zero max", 0, ""),
            ("Negative max", -10, ""),
        ]
    )
    def test_trimmed_values(self, input_str: str, max_length: int, expect_str: str) -> None:
        result = get_trimmed_header(input_str, max_length)
        self.assertEqual(result, expect_str)


class TestToListStr(unittest.TestCase):
    def test_empty_str_returns_empty_list(self) -> None:
        self.assertEqual(to_list_str(""), [])
        self.assertEqual(to_list_str("", sep="|"), [])

    @parameterized.expand(
        [
            ("a,b,c", None, ["a", "b", "c"]),
            ("a,b,c", ",", ["a", "b", "c"]),
            ("a,b,c", ";", ["a,b,c"]),
            ("a,b;c", ";", ["a,b", "c"]),
            ("abc", None, ["abc"]),
            ("abc", ",", ["abc"]),
            ("ab|b", "|", ["ab", "b"]),
        ]
    )
    def test_values(self, input_str: str, sep: str | None, expect_list: list[str]) -> None:
        if sep:
            self.assertEqual(to_list_str(input_str, sep=sep), expect_list)
        else:
            self.assertEqual(to_list_str(input_str), expect_list)



import unittest
from unittest.mock import Mock, patch

from typer.testing import CliRunner

from jupyter_deploy.cli.users_app import users_app


class TestUsersApp(unittest.TestCase):
    """Test cases for the users_app module."""

    def test_help_command(self) -> None:
        self.assertTrue(len(users_app.info.help or "") > 0, "help should not be empty")

        runner = CliRunner()
        result = runner.invoke(users_app, ["--help"])

        self.assertEqual(result.exit_code, 0)
        self.assertTrue(result.stdout.index("add") > 0)
        self.assertTrue(result.stdout.index("remove") > 0)
        self.assertTrue(result.stdout.index("set") > 0)
        self.assertTrue(result.stdout.index("list") > 0)

    def test_no_arg_defaults_to_help(self) -> None:
        runner = CliRunner()
        result = runner.invoke(users_app, [])

        self.assertEqual(result.exit_code, 0)
        self.assertTrue(len(result.stdout) > 0)


class TestUserAddCmd(unittest.TestCase):
    def get_mock_users_handler(self) -> tuple[Mock, dict[str, Mock]]:
        """Return a mock users handler."""
        mock_users_handler = Mock()
        mock_add_users = Mock()
        mock_get_console = Mock()

        mock_users_handler.add_users = mock_add_users
        mock_users_handler.get_console = mock_get_console

        mock_get_console.return_value = Mock()

        return mock_users_handler, {
            "add_users": mock_add_users,
            "get_console": mock_get_console,
        }

    @patch("jupyter_deploy.handlers.access.user_handler.UsersHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_instantiates_users_handler_and_calls_add_users(
        self, mock_project_dir: Mock, mock_users_handler_class: Mock
    ) -> None:
        """Test that add command instantiates UsersHandler and calls add_users."""
        # Setup
        mock_users_handler, mock_handler_fns = self.get_mock_users_handler()
        mock_users_handler_class.return_value = mock_users_handler
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(users_app, ["add", "user1", "user2"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_users_handler_class.assert_called_once()
        mock_handler_fns["add_users"].assert_called_once_with(["user1", "user2"])

    @patch("jupyter_deploy.handlers.access.user_handler.UsersHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_switches_dir_when_passed_a_project(self, mock_project_dir: Mock, mock_users_handler_class: Mock) -> None:
        # Setup
        mock_users_handler, _ = self.get_mock_users_handler()
        mock_users_handler_class.return_value = mock_users_handler
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(users_app, ["add", "user1", "--path", "/test/project/path"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_project_dir.assert_called_once_with("/test/project/path")

    @patch("jupyter_deploy.handlers.access.user_handler.UsersHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_raises_when_users_handler_add_users_raises(
        self, mock_project_dir: Mock, mock_users_handler_class: Mock
    ) -> None:
        """Test that add command propagates exceptions from add_users."""
        # Setup
        mock_users_handler, mock_handler_fns = self.get_mock_users_handler()
        mock_users_handler_class.return_value = mock_users_handler
        mock_handler_fns["add_users"].side_effect = Exception("Test error")
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(users_app, ["add", "user1"])

        # Assert
        self.assertNotEqual(result.exit_code, 0)


class TestUserRemoveCmd(unittest.TestCase):
    def get_mock_users_handler(self) -> tuple[Mock, dict[str, Mock]]:
        """Return a mock users handler."""
        mock_users_handler = Mock()
        mock_remove_users = Mock()
        mock_get_console = Mock()

        mock_users_handler.remove_users = mock_remove_users
        mock_users_handler.get_console = mock_get_console

        mock_get_console.return_value = Mock()

        return mock_users_handler, {
            "remove_users": mock_remove_users,
            "get_console": mock_get_console,
        }

    @patch("jupyter_deploy.handlers.access.user_handler.UsersHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_instantiates_users_handler_and_calls_remove_users(
        self, mock_project_dir: Mock, mock_users_handler_class: Mock
    ) -> None:
        # Setup
        mock_users_handler, mock_handler_fns = self.get_mock_users_handler()
        mock_users_handler_class.return_value = mock_users_handler
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(users_app, ["remove", "user1", "user2"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_users_handler_class.assert_called_once()
        mock_handler_fns["remove_users"].assert_called_once_with(["user1", "user2"])

    @patch("jupyter_deploy.handlers.access.user_handler.UsersHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_switches_dir_when_passed_a_project(self, mock_project_dir: Mock, mock_users_handler_class: Mock) -> None:
        # Setup
        mock_users_handler, _ = self.get_mock_users_handler()
        mock_users_handler_class.return_value = mock_users_handler
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(users_app, ["remove", "user1", "--path", "/test/project/path"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_project_dir.assert_called_once_with("/test/project/path")

    @patch("jupyter_deploy.handlers.access.user_handler.UsersHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_raises_when_users_handler_remove_users_raises(
        self, mock_project_dir: Mock, mock_users_handler_class: Mock
    ) -> None:
        # Setup
        mock_users_handler, mock_handler_fns = self.get_mock_users_handler()
        mock_users_handler_class.return_value = mock_users_handler
        mock_handler_fns["remove_users"].side_effect = Exception("Test error")
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(users_app, ["remove", "user1"])

        # Assert
        self.assertNotEqual(result.exit_code, 0)


class TestUserSetCmd(unittest.TestCase):
    def get_mock_users_handler(self) -> tuple[Mock, dict[str, Mock]]:
        """Return a mock users handler."""
        mock_users_handler = Mock()
        mock_set_users = Mock()
        mock_get_console = Mock()

        mock_users_handler.set_users = mock_set_users
        mock_users_handler.get_console = mock_get_console

        mock_get_console.return_value = Mock()

        return mock_users_handler, {
            "set_users": mock_set_users,
            "get_console": mock_get_console,
        }

    @patch("jupyter_deploy.handlers.access.user_handler.UsersHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_instantiates_users_handler_and_calls_set_users(
        self, mock_project_dir: Mock, mock_users_handler_class: Mock
    ) -> None:
        """Test that set command instantiates UsersHandler and calls set_users."""
        # Setup
        mock_users_handler, mock_handler_fns = self.get_mock_users_handler()
        mock_users_handler_class.return_value = mock_users_handler
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(users_app, ["set", "user1", "user2"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_users_handler_class.assert_called_once()
        mock_handler_fns["set_users"].assert_called_once_with(["user1", "user2"])

    @patch("jupyter_deploy.handlers.access.user_handler.UsersHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_switches_dir_when_passed_a_project(self, mock_project_dir: Mock, mock_users_handler_class: Mock) -> None:
        """Test that set command switches directory when a project path is provided."""
        # Setup
        mock_users_handler, _ = self.get_mock_users_handler()
        mock_users_handler_class.return_value = mock_users_handler
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(users_app, ["set", "user1", "--path", "/test/project/path"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_project_dir.assert_called_once_with("/test/project/path")

    @patch("jupyter_deploy.handlers.access.user_handler.UsersHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_raises_when_users_handler_set_users_raises(
        self, mock_project_dir: Mock, mock_users_handler_class: Mock
    ) -> None:
        """Test that set command propagates exceptions from set_users."""
        # Setup
        mock_users_handler, mock_handler_fns = self.get_mock_users_handler()
        mock_users_handler_class.return_value = mock_users_handler
        mock_handler_fns["set_users"].side_effect = Exception("Test error")
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(users_app, ["set", "user1"])

        # Assert
        self.assertNotEqual(result.exit_code, 0)


class TestUserListCmd(unittest.TestCase):
    def get_mock_users_handler(self) -> tuple[Mock, dict[str, Mock]]:
        """Return a mock users handler."""
        mock_users_handler = Mock()
        mock_list_users = Mock()
        mock_get_console = Mock()

        mock_users_handler.list_users = mock_list_users
        mock_users_handler.get_console = mock_get_console

        mock_list_users.return_value = ["user1", "user2"]
        mock_get_console.return_value = Mock()

        return mock_users_handler, {
            "list_users": mock_list_users,
            "get_console": mock_get_console,
        }

    @patch("jupyter_deploy.handlers.access.user_handler.UsersHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_instantiates_users_handler_and_calls_list_users(
        self, mock_project_dir: Mock, mock_users_handler_class: Mock
    ) -> None:
        """Test that list command instantiates UsersHandler and calls list_users."""
        # Setup
        mock_users_handler, mock_handler_fns = self.get_mock_users_handler()
        mock_users_handler_class.return_value = mock_users_handler
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(users_app, ["list"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_users_handler_class.assert_called_once()
        mock_handler_fns["list_users"].assert_called_once()

    @patch("jupyter_deploy.handlers.access.user_handler.UsersHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_uses_handler_console_to_print_users_list(
        self, mock_project_dir: Mock, mock_users_handler_class: Mock
    ) -> None:
        """Test that list command uses the handler's console to print the list."""
        # Setup
        mock_users_handler, mock_handler_fns = self.get_mock_users_handler()
        mock_users_handler_class.return_value = mock_users_handler

        mock_console = Mock()
        mock_handler_fns["get_console"].return_value = mock_console
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(users_app, ["list"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_console.print.assert_called_once()
        mock_call = mock_console.print.mock_calls[0]
        self.assertTrue("user1, user2" in str(mock_call))

    @patch("jupyter_deploy.handlers.access.user_handler.UsersHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_handles_no_users(self, mock_project_dir: Mock, mock_users_handler_class: Mock) -> None:
        """Test that list command handles the case when no users are allowlisted."""
        # Setup
        mock_users_handler, mock_handler_fns = self.get_mock_users_handler()
        mock_users_handler_class.return_value = mock_users_handler
        mock_handler_fns["list_users"].return_value = []

        mock_console = Mock()
        mock_handler_fns["get_console"].return_value = mock_console
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(users_app, ["list"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_console.print.assert_called_once()
        mock_call = mock_console.print.mock_calls[0]
        self.assertTrue("None" in str(mock_call))

    @patch("jupyter_deploy.handlers.access.user_handler.UsersHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_switches_dir_when_passed_a_project(self, mock_project_dir: Mock, mock_users_handler_class: Mock) -> None:
        """Test that list command switches directory when a project path is provided."""
        # Setup
        mock_users_handler, _ = self.get_mock_users_handler()
        mock_users_handler_class.return_value = mock_users_handler
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(users_app, ["list", "--path", "/test/project/path"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_project_dir.assert_called_once_with("/test/project/path")

    @patch("jupyter_deploy.handlers.access.user_handler.UsersHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_raises_when_users_handler_list_users_raises(
        self, mock_project_dir: Mock, mock_users_handler_class: Mock
    ) -> None:
        """Test that list command propagates exceptions from list_users."""
        # Setup
        mock_users_handler, mock_handler_fns = self.get_mock_users_handler()
        mock_users_handler_class.return_value = mock_users_handler
        mock_handler_fns["list_users"].side_effect = Exception("Test error")
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(users_app, ["list"])

        # Assert
        self.assertNotEqual(result.exit_code, 0)
